from .skyproj import McBrydeSkyproj, LaeaSkyproj, AlbersSkyproj
from .utils import get_datafile  # fix this

from ._docstrings import skyproj_init_parameters, skyproj_kwargs_par
from ._docstrings import (
    add_func_docstr,
    draw_hpxmap_docstr,
    draw_hpxpix_docstr,
    draw_hspmap_docstr,
    draw_hpxbin_docstr,
)

__all__ = ['DESSkyproj', 'BlissSkyproj', 'MaglitesSkyproj',
           'DecalsSkyproj', 'DESMcBrydeSkyproj', 'DESAlbersSkyproj']


class _Survey:
    """Extension class to add routines for drawing survey outlines.
    """
    def draw_des(self, **kwargs):
        """Draw the DES footprint."""
        return self.draw_des19(**kwargs)

    def draw_des19(self, edgecolor='red', lw=2, **kwargs):
        """Draw the DES 2019 footprint."""
        filename = get_datafile('des-round19-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_des17(self, edgecolor='red', lw=2, **kwargs):
        """Draw the DES 2017 footprint."""
        filename = get_datafile('des-round17-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_decals(self, edgecolor='red', lw=2, **kwargs):
        """Draw the DECaLS footprint."""
        filename = get_datafile('decals-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_maglites(self, edgecolor='blue', lw=2, **kwargs):
        """Draw the MagLiteS footprint."""
        filename = get_datafile('maglites-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_bliss(self, edgecolor='magenta', lw=2, **kwargs):
        """Draw the BLISS footprint."""
        filename = get_datafile('bliss-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    # Override zoom default for survey maps to keep the default fixed.
    @add_func_docstr(draw_hpxmap_docstr)
    def draw_hpxmap(self, hpxmap, nest=False, zoom=False, xsize=1000, vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        return super().draw_hpxmap(hpxmap,
                                   nest=nest,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)

    @add_func_docstr(draw_hpxpix_docstr)
    def draw_hpxpix(self, nside, pixels, values, nest=False, zoom=False, xsize=1000,
                    vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        return super().draw_hpxpix(nside,
                                   pixels,
                                   values,
                                   nest=nest,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)

    @add_func_docstr(draw_hspmap_docstr)
    def draw_hspmap(self, hspmap, zoom=False, xsize=1000, vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        return super().draw_hspmap(hspmap,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)

    @add_func_docstr(draw_hpxbin_docstr)
    def draw_hpxbin(self, lon, lat, C=None, nside=256, nest=False, zoom=False, xsize=1000,
                    vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        return super().draw_hpxbin(lon,
                                   lat,
                                   C=C,
                                   nside=nside,
                                   nest=nest,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)


class DESMcBrydeSkyproj(_Survey, McBrydeSkyproj):
    __doc__ = skyproj_init_parameters(
        "A projection for the DES Survey using a McBryde-Thomas Flat Polar Quartic projection.",
    )
    __doc__ += skyproj_kwargs_par

    def __init__(
        self,
        ax=None,
        *,
        lon_0=30.0,
        gridlines=True,
        celestial=True,
        extent=[90, -50, -74, 10],
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=10,
        n_grid_lat=6,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )


DESSkyproj = DESMcBrydeSkyproj


class DESAlbersSkyproj(_Survey, AlbersSkyproj):
    __doc__ = skyproj_init_parameters(
        "A projection for the DES Survey using an Albers Equal Area projection.",
        include_lon_0=False,
    )
    __doc__ += skyproj_kwargs_par

    def __init__(
        self,
        ax=None,
        *,
        gridlines=True,
        celestial=True,
        extent=[80, -40, -80, 10],
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=10,
        n_grid_lat=6,
        min_lon_ticklabel_delta=0.1,
        lon_0=30.0,
        lat_1=-15.0,
        lat_2=-50.0,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            lon_0=lon_0,
            lat_1=lat_1,
            lat_2=lat_2,
            **kwargs,
        )

    @property
    def _default_xy_labels(self):
        return ("Right Ascension", "Declination")


class BlissSkyproj(_Survey, McBrydeSkyproj):
    __doc__ = skyproj_init_parameters(
        "A projection for the BLISS Survey using a McBryde-Thomas Flat Polar Quartic projection.",
        include_lon_0=False,
    )
    __doc__ += skyproj_kwargs_par

    def __init__(
        self,
        ax=None,
        *,
        lon_0=100.0,
        gridlines=True,
        celestial=True,
        extent=[-60, 250, -55, 0],
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=10,
        n_grid_lat=6,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )


class MaglitesSkyproj(_Survey, LaeaSkyproj):
    __doc__ = skyproj_init_parameters(
        "A projection for the MagLiteS Survey using a Lambert Azimuthal Equal Area projection.",
        include_lon_0=False,
    )
    __doc__ += skyproj_kwargs_par

    def __init__(
        self,
        ax=None,
        *,
        gridlines=True,
        celestial=True,
        extent=[-150, 70, -85, -50],
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=10,
        n_grid_lat=6,
        min_lon_ticklabel_delta=0.1,
        lon_0=0.0,
        lat_0=-90.0,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            lon_0=lon_0,
            lat_0=lat_0,
            **kwargs,
        )


class DecalsSkyproj(_Survey, McBrydeSkyproj):
    __doc__ = skyproj_init_parameters(
        "A projection for the DECaLS Survey using a McBryde-Thomas Flat Polar Quartic projection.",
        include_lon_0=False,
    )
    __doc__ += skyproj_kwargs_par

    def __init__(
        self,
        ax=None,
        *,
        lon_0=105.0,
        gridlines=True,
        celestial=True,
        extent=[180, -180, -30, 40],
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=10,
        n_grid_lat=6,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )



import numpy as np
from packaging import version

from mpl_toolkits.axisartist.grid_finder import ExtremeFinderSimple
import mpl_toolkits.axisartist.angle_helper as angle_helper
from mpl_toolkits.axisartist.grid_helper_curvelinear import GridHelperCurveLinear

from .utils import wrap_values

import matplotlib
if version.parse(matplotlib.__version__) >= version.parse('3.5'):
    _celestial_angle_360 = True
else:
    _celestial_angle_360 = False

__all__ = ['WrappedFormatterDMS', 'ExtremeFinderWrapped', 'GridHelperSkyproj']


class WrappedFormatterDMS(angle_helper.FormatterDMS):
    """A tick formatter that handles longitude wrapping.

    Parameters
    ----------
    wrap : `float`
        Angle (degrees) which should be wrapped (subtract 360).
    longitude_ticks : `str`
        Type of longitude ticks, either ``positive`` (0 to 360)
        or ``symmetric`` (-180 to 180).
    """
    def __init__(self, wrap, longitude_ticks):
        self._wrap = wrap
        if longitude_ticks == 'positive':
            self._longitude_ticks = 1
        elif longitude_ticks == 'symmetric':
            self._longitude_ticks = -1
        else:
            raise ValueError("longitude_ticks must be `positive` or `symmetric`.")
        super().__init__()

    def _wrap_values(self, factor, values):
        """Wrap the values according to the wrap angle.

        Parameters
        ----------
        factor : `float`
            Scaling factor for input values
        values : `list`
            List of values to format

        Returns
        -------
        wrapped_values : `np.ndarray`
            Array of wrapped values, scaled by factor.
        """
        _values = np.atleast_1d(values)/factor
        _values = wrap_values(_values, wrap=self._wrap)
        if self._longitude_ticks == 1:
            # Values should all be positive, 0 to 360
            _values %= 360.0
        else:
            # Additional test to set -180 to positive 180
            _values[np.isclose(_values, -180.0)] = 180.0
        return factor*_values

    def __call__(self, direction, factor, values):
        return super().__call__(direction, factor, self._wrap_values(factor, values))


class ExtremeFinderWrapped(ExtremeFinderSimple):
    """
    Find extremes with configurable wrap angle and correct limits.

    Parameters
    ----------
    nx : `int`
        Number of samples in x direction.
    ny : `int`
        Number of samples in y direction.
    wrap_angle : `float`
        Angle at which the 360-degree cycle should be wrapped.
    """
    def __init__(self, nx, ny, wrap_angle):
        self.nx, self.ny = nx, ny
        self._wrap = wrap_angle
        self._eps = 1e-5

    def __call__(self, transform_xy, x1, y1, x2, y2):
        # docstring inherited
        x, y = np.meshgrid(
            np.linspace(x1, x2, self.nx), np.linspace(y1, y2, self.ny))
        lon, lat = transform_xy(np.ravel(x), np.ravel(y))

        with np.errstate(invalid='ignore'):
            lon = wrap_values(lon, wrap=self._wrap)

        lon_min, lon_max = np.nanmin(lon), np.nanmax(lon)
        lat_min, lat_max = np.nanmin(lat), np.nanmax(lat)

        lon_min, lon_max, lat_min, lat_max = \
            self._add_pad(lon_min, lon_max, lat_min, lat_max)

        lat_min = np.clip(lat_min, -90.0 + self._eps, 90.0 - self._eps)
        lat_max = np.clip(lat_max, -90.0 + self._eps, 90.0 - self._eps)

        lon_min = np.clip(lon_min, self._wrap - 360. + self._eps, self._wrap - self._eps)
        lon_max = np.clip(lon_max, self._wrap - 360. + self._eps, self._wrap - self._eps)

        return lon_min, lon_max, lat_min, lat_max


class GridHelperSkyproj(GridHelperCurveLinear):
    """GridHelperCurveLinear with tick overlap protection.

    Parameters
    ----------
    *args : `list`
        Arguments for ``GridHelperCurveLinear``.
    celestial : `bool`, optional
        Plot is celestial, and angles should be 0 to 360.  Otherwise -180 to 180.
    equatorial_labels : `bool`, optional
        Longitude labels are marked on the equator instead of edges.
    delta_cut : `float`, optional
        Gridline step (degrees) to signify a jump around a wrapped edge.
    min_lon_ticklabel_delta : `float`, optional
        Minimum relative spacing between longitude tick labels (relative to width
        of axis). Smaller values yield closer tick labels (and potential for clashes)
        and larger values yield more spacing between tick labels.
    **kwargs : `dict`, optional
        Additional kwargs for ``GridHelperCurveLinear``.
    """
    def __init__(self, *args, celestial=True, equatorial_labels=False, delta_cut=80.0,
                 min_lon_ticklabel_delta=0.1, **kwargs):
        self._celestial = celestial
        self._equatorial_labels = equatorial_labels
        self._delta_cut = delta_cut
        self._min_lon_ticklabel_delta = min_lon_ticklabel_delta

        super().__init__(*args, **kwargs)

    def get_gridlines(self, which="major", axis="both"):
        # docstring inherited
        try:
            _grid_info = self._grid_info
        except AttributeError:
            _grid_info = self.grid_info

        grid_lines = []
        if axis in ["both", "x"]:
            for gl in _grid_info["lon"]["lines"]:
                grid_lines.extend(self._cut_grid_line_jumps(gl))
        if axis in ["both", "y"]:
            for gl in _grid_info["lat"]["lines"]:
                grid_lines.extend(self._cut_grid_line_jumps(gl))
        return grid_lines

    def _cut_grid_line_jumps(self, gl):
        """Check for jumps and cut gridlines into multiple sections.

        Parameters
        ----------
        gl : `list` [`tuple`]
            Input gridlines.  List of tuples of numpy arrays.

        Returns
        -------
        gl_new : `list` [`tuple`]
            New gridlines.  Jumps have been replaced with `np.nan`
            values to ensure lines are not connected around edges.
        """
        dx = gl[0][0][1:] - gl[0][0][: -1]
        dy = gl[0][1][1:] - gl[0][1][: -1]

        split, = (np.hypot(dx, dy) > self._delta_cut).nonzero()

        if split.size == 0:
            return gl

        gl_new = [(np.insert(gl[0][0], split + 1, np.nan),
                   np.insert(gl[0][1], split + 1, np.nan))]

        return gl_new

    def get_tick_iterator(self, nth_coord, axis_side, minor=False):
        # docstring inherited
        try:
            _grid_info = self._grid_info
        except AttributeError:
            _grid_info = self.grid_info

        angle_tangent = dict(left=90, right=90, bottom=0, top=0)[axis_side]
        lon_or_lat = ["lon", "lat"][nth_coord]
        if lon_or_lat == "lon":
            # Need to compute maximum extent in the x direction
            x_min = 1e100
            x_max = -1e100
            for line in _grid_info['lon_lines']:
                x_min = min((x_min, np.min(line[0])))
                x_max = max((x_max, np.max(line[0])))
            delta_x = x_max - x_min
        if not minor:  # major ticks
            tick_locs = _grid_info[lon_or_lat]["tick_locs"][axis_side]
            tick_labels = _grid_info[lon_or_lat]["tick_labels"][axis_side]

            if lon_or_lat == "lon" and self._equatorial_labels:
                min_lat = np.min(_grid_info["lat"]["levels"])
                max_lat = np.max(_grid_info["lat"]["levels"])
                if (min_lat < -89.0 and axis_side == "bottom") \
                   or (max_lat > 89.0 and axis_side == "top"):
                    tick_locs = []
                    tick_labels = []

            prev_xy = None
            for ctr, ((xy, a), l) in enumerate(zip(tick_locs, tick_labels)):
                if self._celestial and _celestial_angle_360:
                    angle_normal = 360.0 - a
                else:
                    angle_normal = a

                if ctr > 0 and lon_or_lat == 'lon':
                    # Check if this is too close to the last label.
                    if abs(xy[0] - prev_xy[0])/delta_x < self._min_lon_ticklabel_delta:
                        continue
                prev_xy = xy
                yield xy, angle_normal, angle_tangent, l
        else:
            for (xy, a), l in zip(
                    _grid_info[lon_or_lat]["tick_locs"][axis_side],
                    _grid_info[lon_or_lat]["tick_labels"][axis_side]):
                if self._celestial and _celestial_angle_360:
                    angle_normal = 360.0 - a
                else:
                    angle_normal = a
                yield xy, angle_normal, angle_tangent, ""



import warnings

import matplotlib.pyplot as plt

import numpy as np
import hpgeom as hpg

import mpl_toolkits.axisartist as axisartist
import mpl_toolkits.axisartist.angle_helper as angle_helper
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize

from .skycrs import get_crs, PlateCarreeCRS, GnomonicCRS
from .hpx_utils import healpix_pixels_range, hspmap_to_xy, hpxmap_to_xy, healpix_to_xy, healpix_bin
from .mpl_utils import ExtremeFinderWrapped, WrappedFormatterDMS, GridHelperSkyproj
from .utils import wrap_values, _get_boundary_poly_xy, get_autoscale_vmin_vmax

from ._docstrings import skyproj_init_parameters, skyproj_kwargs_par
from ._docstrings import (
    add_func_docstr,
    draw_hpxmap_docstr,
    draw_hpxpix_docstr,
    draw_hspmap_docstr,
    draw_hpxbin_docstr,
)


class _Skyproj():
    __doc__ = skyproj_init_parameters(
        "Base class for creating Skyproj objects.",
        include_projection_name=True,
    ) + skyproj_kwargs_par

    def __init__(
        self,
        ax=None,
        *,
        projection_name='cyl',
        lon_0=0.0,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        self._redraw_dict = {'hpxmap': None,
                             'hspmap': None,
                             'im': None,
                             'inset_colorbar': None,
                             'inset_colorbar_kwargs': {},
                             'colorbar': None,
                             'colorbar_kwargs': {},
                             'lon_range_home': None,
                             'lat_range_home': None,
                             'vmin': None,
                             'vmax': None,
                             'xsize': None,
                             'kwargs_pcolormesh': None,
                             'nside': None,
                             'nest': None,
                             'rasterized': True}

        if longitude_ticks == 'positive':
            self._longitude_ticks = longitude_ticks
        elif longitude_ticks == 'symmetric':
            self._longitude_ticks = longitude_ticks
        else:
            raise ValueError("longitude_ticks must be 'positive' or 'symmetric'.")

        self._n_grid_lon = n_grid_lon
        self._n_grid_lat = n_grid_lat
        self._min_lon_ticklabel_delta = min_lon_ticklabel_delta

        if ax is None:
            ax = plt.gca()

        fig = ax.figure
        # This code does not work with the constrained_layout option
        try:
            # Newer matplotlib
            fig.set_layout_engine('none')
        except AttributeError:
            # Older matplotlib
            fig.set_constrained_layout(False)
        subspec = ax.get_subplotspec()
        fig.delaxes(ax)

        # Map lon_0 to be between -180.0 and 180.0
        lon_0 = wrap_values(lon_0)

        if abs(lon_0) == 180.0:
            # We must move this by epsilon or the code gets confused with 0 == 360
            lon_0 = 179.9999

        kwargs['lon_0'] = lon_0
        crs = get_crs(projection_name, **kwargs)
        self._ax = fig.add_subplot(subspec, projection=crs)
        self._crs_orig = crs
        self._reprojected = False

        self._aa = None

        self.do_celestial = celestial
        self.do_gridlines = gridlines
        self._autorescale = autorescale
        self._galactic = galactic

        self._wrap = (self.lon_0 + 180.) % 360.

        extent_xy = None
        if extent is None:
            extent = self._full_sky_extent_initial
            if self._init_extent_xy:
                # Certain projections (such as laea) require us to know the x/y extent
                proj_boundary_xy = self._compute_proj_boundary_xy()
                pts = np.concatenate((proj_boundary_xy['left'],
                                      proj_boundary_xy['right'],
                                      proj_boundary_xy['top'],
                                      proj_boundary_xy['bottom']))
                extent_xy = [np.min(pts[:, 0]), np.max(pts[:, 0]),
                             np.min(pts[:, 1]), np.max(pts[:, 1])]
        else:
            extent_xy = None

        self._boundary_lines = None
        self._boundary_labels = []

        if "axes.labelsize" not in rcparams:
            rcparams["axes.labelsize"] = 16

        with plt.rc_context(rcparams):
            self._initialize_axes(extent, extent_xy=extent_xy)

            # Set up callbacks on axis zoom.
            self._add_change_axis_callbacks()

            # Set up callback on figure resize.
            self._frc = self.ax.figure.canvas.mpl_connect('resize_event', self._change_size)
            self._dc = self.ax.figure.canvas.mpl_connect('draw_event', self._draw_callback)
            self._initial_extent_xy = [0]*4

            # Set up reproject callback.
            self._rpc = self.ax.figure.canvas.mpl_connect('key_press_event', self._keypress_callback)

            self._draw_aa_bounds_and_labels()

    def proj(self, lon, lat):
        """Apply forward projection to a set of lon/lat positions.

        Convert from lon/lat to x/y.

        Parameters
        ----------
        lon : `float` or `list` or `np.ndarray`
            Array of longitude(s) (degrees).
        lat : `float` or `list` or `np.ndarray`
            Array of latitudes(s) (degrees).

        Returns
        -------
        x : `np.ndarray`
            Array of x values.
        y : `np.ndarray`
            Array of y values.
        """
        lon = np.atleast_1d(lon)
        lat = np.atleast_1d(lat)
        out = ((lat < (-90.0 + self._pole_clip))
               | (lat > (90.0 - self._pole_clip)))
        proj_xy = self.crs.transform_points(PlateCarreeCRS(), lon, lat)
        # FIXME I don't like this, look at the get_extent code instead/as well?
        proj_xy[..., 1][out] = np.nan
        return proj_xy[..., 0], proj_xy[..., 1]

    def proj_inverse(self, x, y):
        """Apply inverse projection to a set of points.

        Convert from x/y to lon/lat.

        Parameters
        ----------
        x : `float` or `list` or `np.ndarray`
            Projected x values.
        y : `float` or `list` or `np.ndarray`
            Projected y values.

        Returns
        -------
        lon : `np.ndarray`
            Array of longitudes (degrees).
        lat : `np.ndarray`
            Array of latitudes (degrees).
        """
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        proj_lonlat = PlateCarreeCRS().transform_points(self.crs, x, y)
        return proj_lonlat[..., 0], proj_lonlat[..., 1]

    def _initialize_axes(self, extent, extent_xy=None):
        """Initialize the axes with a given extent.

        Note that calling this method will remove all formatting options.

        Parameters
        ----------
        extent : array-like
            Extent as [lon_min, lon_max, lat_min, lat_max].
        extent_xy : array-like, optional
            Extent in xy space [x_min, x_max, y_min, y_max].
            Used for full-sky initialization.
        """
        # Reset any axis artist if necessary
        if self._aa is not None:
            self._aa.remove()
            self._aa = None

        self._set_axes_limits(extent, extent_xy=extent_xy, invert=False)
        self._create_axes(extent)
        self._set_axes_limits(extent, extent_xy=extent_xy, invert=self.do_celestial)

        self._ax.set_frame_on(False)
        if self.do_gridlines:
            self._aa.grid(True, linestyle=':', color='k', lw=0.5)

        self._aa.axis[:].line.set_visible(False)
        self._aa.axis[:].major_ticks.set_visible(False)

        self._extent_xy = self._ax.get_extent(lonlat=False)
        self._changed_x_axis = False
        self._changed_y_axis = False

    def set_extent(self, extent):
        """Set the extent.

        Axes will be properly inverted if Skyproj was initialized with
        ``celestial=True``.

        Parameters
        ----------
        extent : array-like
            Extent as [lon_min, lon_max, lat_min, lat_max].
        """
        self._set_axes_limits(extent, invert=self.do_celestial)
        self._extent_xy = self._ax.get_extent(lonlat=False)

        if self._n_grid_lon is None or self._n_grid_lat is None:
            n_grid_lon, n_grid_lat = self._compute_n_grid_from_extent(extent)

            self._update_grid_locators(n_grid_lon, n_grid_lat)

        self._draw_aa_bounds_and_labels()

    def _compute_n_grid_from_extent(self, extent):
        """Compute the number of grid lines from the extent.

        This will respect values that were set at initialization time.

        Parameters
        ----------
        extent : array-like
            Extent as [lon_min, lon_max, lat_min, lat_max].

        Returns
        -------
        n_grid_lon : `int`
            Number of gridlines in the longitude direction.
        n_grid_lat : `int`
            Number of gridlines in the latitude direction.
        """
        n_grid_lat = self._n_grid_lat if self._n_grid_lat is not None else 6
        if self._n_grid_lon is None:
            latscale = np.cos(np.deg2rad(np.mean(extent[2:])))
            ratio = np.clip(np.abs(extent[1] - extent[0])*latscale/(extent[3] - extent[2]), 1./3., 5./3.)
            n_grid_lon = int(np.ceil(ratio * n_grid_lat))
        else:
            n_grid_lon = self._n_grid_lon

        return n_grid_lon, n_grid_lat

    def _update_grid_locators(self, n_grid_lon, n_grid_lat):
        """Update the grid locators.

        Parameters
        ----------
        n_grid_lon : `int`
            Number of gridlines in the longitude direction.
        n_grid_lat : `int`
            Number of gridlines in the latitude direction.
        """
        if self._wrap == 180.0 and not self._full_circle:
            include_last_lon = True
        else:
            include_last_lon = False

        if n_grid_lon != self._grid_helper.grid_finder.grid_locator1.nbins:
            self._grid_helper.grid_finder.grid_locator1 = angle_helper.LocatorD(
                n_grid_lon,
                include_last=include_last_lon,
            )
        if n_grid_lat != self._grid_helper.grid_finder.grid_locator2.nbins:
            self._grid_helper.grid_finder.grid_locator2 = angle_helper.LocatorD(
                n_grid_lat,
                include_last=True,
            )

    def _draw_aa_bounds_and_labels(self):
        """Set the axisartist bounds and labels."""
        # Remove any previous lines
        if self._boundary_lines:
            for line in self._boundary_lines:
                line.remove()
            self._boundary_lines = None

        extent_xy = self._ax.get_extent(lonlat=False)
        bounds_xy = self._compute_proj_boundary_xy()
        bounds_xy_clipped = _get_boundary_poly_xy(bounds_xy, extent_xy, self.proj, self.proj_inverse)

        self._boundary_lines = self._ax.plot(bounds_xy_clipped[:, 0],
                                             bounds_xy_clipped[:, 1],
                                             'k-',
                                             lonlat=False,
                                             clip_on=False,
                                             linewidth=plt.rcParams['axes.linewidth'])

        self._aa.axis[:].line.set_visible(False)
        self._aa.axis[:].major_ticks.set_visible(False)

        # Remove any previous labels
        if self._boundary_labels:
            for label in self._boundary_labels:
                label.remove()
            self._boundary_labels = []

        grid_finder = self._grid_helper.grid_finder
        grid_info = grid_finder.get_grid_info(
            extent_xy[0],
            extent_xy[2],
            extent_xy[1],
            extent_xy[3]
        )

        # Recover lon_level, lat_level
        extremes = grid_finder.extreme_finder(
            grid_finder.inv_transform_xy,
            extent_xy[0],
            extent_xy[2],
            extent_xy[1],
            extent_xy[3]
        )
        _, _, lon_factor = grid_finder.grid_locator1(extremes[0], extremes[1])
        _, _, lat_factor = grid_finder.grid_locator2(extremes[2], extremes[3])

        self._boundary_labels.extend(self._draw_aa_lat_labels(extent_xy, grid_info, lat_factor))
        self._boundary_labels.extend(self._draw_aa_lon_labels(extent_xy, grid_info, lon_factor))

    def _draw_aa_lat_labels(self, extent_xy, grid_info, factor):
        """Draw axis artist latitude labels.

        Parameters
        ----------
        extent_xy : `list`
            Extent in x/y space
        grid_info : `dict`
            Grid info to determine label locations
        factor : `float`
            Multiplicative factor to convert ticks to values.

        Returns
        -------
        labels : `list` [`matplotlib.Text`]
        """
        levels = grid_info['lat']['levels']
        lines = grid_info['lat']['lines']

        inverted = (extent_xy[1] < extent_xy[0])

        # The grid_info will be reversed left/right if the axis is inverted.
        if inverted:
            gi_side_map = {'left': 'right',
                           'right': 'left'}
            x0_index = 1
            x1_index = 0
        else:
            gi_side_map = {side: side for side in ['left', 'right']}
            x0_index = 0
            x1_index = 1

        boundary_labels = []

        for axis_side in ['left', 'right']:
            if not self._aa.axis[axis_side].major_ticklabels.get_visible():
                continue

            tick_levels = grid_info['lat']['tick_levels'][gi_side_map[axis_side]]

            for lat_level, lat_line in zip(levels, lines):
                if np.abs(np.abs(lat_level) - 90.0) < 1.0:
                    continue

                if int(lat_level*factor) in tick_levels:
                    continue

                lat_line_x = lat_line[0][0]
                lat_line_y = lat_line[0][1]

                if gi_side_map[axis_side] == 'right':
                    lat_line_x = lat_line_x[::-1]
                    lat_line_y = lat_line_y[::-1]

                if axis_side == 'left':
                    ha = 'right'
                else:
                    ha = 'left'

                if lat_level < 0.0:
                    va = 'top'
                else:
                    va = 'bottom'

                # Skip any that are out of the y bounding box.
                if lat_line_y[0] < extent_xy[2] or lat_line_y[0] > extent_xy[3]:
                    continue

                if lat_line_x[0] < extent_xy[x0_index] or lat_line_y[0] > extent_xy[x1_index]:
                    continue

                label = self._tick_formatter2(axis_side, factor, [lat_level])[0]

                boundary_labels.append(self._ax.text(lat_line_x[0],
                                                     lat_line_y[0],
                                                     label,
                                                     size=plt.rcParams['ytick.labelsize'],
                                                     lonlat=False,
                                                     clip_on=False,
                                                     ha=ha,
                                                     va=va))
        return boundary_labels

    def _draw_aa_lon_labels(self, extent_xy, grid_info, factor):
        """Draw axis artist latitude labels.

        Parameters
        ----------
        extent_xy : `list`
            Extent in x/y space
        grid_info : `dict`
            Grid info to determine label locations
        factor : `float`
            Multiplicative factor to convert ticks to values.

        Returns
        -------
        labels : `list` [`matplotlib.Text`]
        """
        levels = grid_info['lon']['levels']
        lines = grid_info['lon']['lines']

        # Need to compute maximum extent in the x direction
        x_min = 1e100
        x_max = -1e100
        for line in grid_info['lon_lines']:
            x_min = min((x_min, np.min(line[0])))
            x_max = max((x_max, np.max(line[0])))
        delta_x = x_max - x_min

        # The grid_info will be reversed left/right if the axis is inverted.
        inverted = (extent_xy[1] < extent_xy[0])
        if inverted:
            x0_index = 1
            x1_index = 0
        else:
            x0_index = 0
            x1_index = 1

        boundary_labels = []

        draw_equatorial_labels = False
        if self._equatorial_labels:
            min_lat = np.min(grid_info['lat']['levels'])
            max_lat = np.max(grid_info['lat']['levels'])
            if min_lat < -89.0 and max_lat > 89.0:
                draw_equatorial_labels = True

        if draw_equatorial_labels:
            levels = np.array(grid_info['lon']['levels'])

            x, y = self.proj(levels, np.zeros(len(levels)))

            ok, = np.where((x > extent_xy[x0_index]) & (x < extent_xy[x1_index])
                           & (y > extent_xy[2]) & (y < extent_xy[3]))

            prev_x = None
            for i in ok:
                if prev_x is not None:
                    # Check if too close to last label.
                    if abs(x[i] - prev_x)/delta_x < self._min_lon_ticklabel_delta:
                        continue
                prev_x = x[i]

                label = self._tick_formatter1('top', factor, [levels[i]])[0]

                boundary_labels.append(self._ax.text(x[i],
                                                     y[i],
                                                     label,
                                                     size=plt.rcParams['xtick.labelsize'],
                                                     lonlat=False,
                                                     clip_on=False,
                                                     ha='right',
                                                     va='bottom'))
        else:
            for axis_side in ['top', 'bottom']:
                if not self._aa.axis[axis_side].major_ticklabels.get_visible():
                    continue

                if self._radial_labels and axis_side == 'bottom':
                    continue

                if axis_side == 'top':
                    line_index = -1
                else:
                    line_index = 0

                tick_levels = grid_info['lon']['tick_levels'][axis_side]

                prev_x = None
                for lon_level, lon_line in zip(levels, lines):
                    if int(lon_level*factor) in tick_levels:
                        continue

                    lon_line_x = lon_line[0][0]
                    lon_line_y = lon_line[0][1]

                    if lon_line_x[line_index] < extent_xy[x0_index] or \
                       lon_line_x[line_index] > extent_xy[x1_index] \
                       or lon_line_y[line_index] < extent_xy[2] or \
                       lon_line_y[line_index] > extent_xy[3]:
                        continue

                    if axis_side == 'top':
                        va = 'bottom'
                        index = -1
                        y_offset = 0.02*(lon_line_y[-1] - lon_line_y[0])
                    else:
                        va = 'top'
                        index = 0
                        y_offset = -0.02*(lon_line_y[-1] - lon_line_y[0])

                    if prev_x is not None:
                        # check if too close to last label.
                        if abs(lon_line_x[index] - prev_x)/delta_x < self._min_lon_ticklabel_delta:
                            continue

                    prev_x = lon_line_x[index]

                    label = self._tick_formatter1(axis_side, factor, [lon_level])[0]
                    boundary_labels.append(self._ax.text(lon_line_x[index],
                                                         lon_line_y[index] + y_offset,
                                                         label,
                                                         size=plt.rcParams['xtick.labelsize'],
                                                         lonlat=False,
                                                         clip_on=False,
                                                         ha='center',
                                                         va=va))

        return boundary_labels

    def get_extent(self):
        """Get the extent in lon/lat coordinates.

        Returns
        -------
        extent : `list`
            Extent as [lon_min, lon_max, lat_min, lat_max].
        """
        return self._ax.get_extent(lonlat=True)

    def set_autorescale(self, autorescale):
        """Set automatic rescaling after zoom.

        Parameters
        ----------
        autorescale : `bool`
            Automatically rescale after zoom?
        """
        self._autorescale = autorescale

    def _set_axes_limits(self, extent, extent_xy=None, invert=True):
        """Set axis limits from an extent.

        Parameters
        ----------
        extent : array-like
            Extent as [lon_min, lon_max, lat_min, lat_max].
        extent_xy : array-like, optional
            Extent in xy space [x_min, x_max, y_min, y_max].
            Used for full-sky initialization in place of extent.
        """
        if len(extent) != 4:
            raise ValueError("Must specify extent as a 4-element array.")
        if extent_xy:
            if len(extent_xy) != 4:
                raise ValueError("Must specify extent_xy as a 4-element array.")

        if extent_xy:
            self._ax.set_extent(extent_xy, lonlat=False)
        else:
            self._ax.set_extent(extent, lonlat=True)

        if self._aa is not None:
            self._aa.set_xlim(self._ax.get_xlim())
            self._aa.set_ylim(self._ax.get_ylim())

        if invert:
            self._ax.invert_xaxis()
            if self._aa is not None:
                self._aa.invert_xaxis()

        return self._ax.get_xlim(), self._ax.get_ylim()

    def _create_axes(self, extent):
        """Create axes and axis artist.

        Parameters
        ----------
        extent : `list`
            Axis extent [lon_min, lon_max, lat_min, lat_max] (degrees).
        """
        extreme_finder = ExtremeFinderWrapped(20, 20, self._wrap)
        if self._wrap == 180.0 and not self._full_circle:
            include_last_lon = True
        else:
            include_last_lon = False

        n_grid_lon, n_grid_lat = self._compute_n_grid_from_extent(extent)

        grid_locator1 = angle_helper.LocatorD(n_grid_lon, include_last=include_last_lon)
        grid_locator2 = angle_helper.LocatorD(n_grid_lat, include_last=True)

        # We always want the formatting to be wrapped at 180 (-180 to 180)
        self._tick_formatter1 = WrappedFormatterDMS(180.0, self._longitude_ticks)
        self._tick_formatter2 = angle_helper.FormatterDMS()

        def proj_wrap(lon, lat):
            lon = np.atleast_1d(lon)
            lat = np.atleast_1d(lat)
            lon[np.isclose(lon, self._wrap)] = self._wrap - 1e-10
            proj_xy = self.crs.transform_points(PlateCarreeCRS(), lon, lat)
            return proj_xy[..., 0], proj_xy[..., 1]

        if self.crs.name == 'cyl':
            delta_cut = 80.0
        else:
            delta_cut = 0.5*self.crs.radius

        grid_helper = GridHelperSkyproj(
            (proj_wrap, self.proj_inverse),
            extreme_finder=extreme_finder,
            grid_locator1=grid_locator1,
            grid_locator2=grid_locator2,
            tick_formatter1=self._tick_formatter1,
            tick_formatter2=self._tick_formatter2,
            celestial=self.do_celestial,
            equatorial_labels=self._equatorial_labels,
            delta_cut=delta_cut,
            min_lon_ticklabel_delta=self._min_lon_ticklabel_delta,
        )

        self._grid_helper = grid_helper

        fig = self._ax.figure
        rect = self._ax.get_position()
        self._aa = axisartist.Axes(fig, rect, grid_helper=grid_helper, frameon=False, aspect=1.0)
        fig.add_axes(self._aa)

        self._aa.format_coord = self._format_coord
        self._aa.axis['left'].major_ticklabels.set_visible(True)
        self._aa.axis['right'].major_ticklabels.set_visible(False)
        self._aa.axis['bottom'].major_ticklabels.set_visible(True)
        self._aa.axis['top'].major_ticklabels.set_visible(True)

        self.set_xlabel(self._default_xy_labels[0])
        self.set_ylabel(self._default_xy_labels[1])

        fig.sca(self._ax)

        return fig, self._ax

    def _format_coord(self, x, y):
        """Return a coordinate format string.

        Parameters
        ----------
        x : `float`
            x position in projected coords.
        y : `float`
            y position in projected coords.

        Returns
        -------
        coord_string : `str`
            Formatted string.
        """
        lon, lat = self.proj_inverse(x, y)
        # Check out-of-bounds with reversibility (unless already out-of-bounds)
        if np.isfinite(lon) and np.isfinite(lat):
            xx, yy = self.proj(lon, lat)
            if not np.isclose(x, xx) or not np.isclose(y, yy):
                lon, lat = np.nan, np.nan
        if self._longitude_ticks == 'positive' and np.isfinite(lon):
            lon %= 360.0
        coord_string = 'lon=%.6f, lat=%.6f' % (lon, lat)
        if np.isnan(lon) or np.isnan(lat):
            val = hpg.UNSEEN
        elif self._redraw_dict['hspmap'] is not None:
            val = self._redraw_dict['hspmap'].get_values_pos(lon, lat)
        elif self._redraw_dict['hpxmap'] is not None:
            pix = hpg.angle_to_pixel(self._redraw_dict['nside'],
                                     lon,
                                     lat,
                                     nest=self._redraw_dict['nest'])
            val = self._redraw_dict['hpxmap'][pix]
        else:
            return coord_string

        if np.isclose(val, hpg.UNSEEN):
            coord_string += ', val=UNSEEN'
        else:
            coord_string += ', val=%f' % (val)
        return coord_string

    def _add_change_axis_callbacks(self):
        """Add callbacks to change axis."""
        self._xlc = self._ax.callbacks.connect('xlim_changed', self._change_axis)
        self._ylc = self._ax.callbacks.connect('ylim_changed', self._change_axis)

    def _remove_change_axis_callbacks(self):
        """Remove callbacks to change axis."""
        self._ax.callbacks.disconnect(self._xlc)
        self._ax.callbacks.disconnect(self._ylc)

    def _change_axis(self, ax):
        """Callback for axis change.

        Parameters
        ----------
        ax : `skyproj.SkyAxesSubplot`
        """
        extent_xy = ax.get_extent(lonlat=False)
        if not np.isclose(extent_xy[0], self._extent_xy[0]) \
           or not np.isclose(extent_xy[1], self._extent_xy[1]):
            self._changed_x_axis = True
        if not np.isclose(extent_xy[2], self._extent_xy[2]) or \
           not np.isclose(extent_xy[3], self._extent_xy[3]):
            self._changed_y_axis = True

        if not self._changed_x_axis or not self._changed_y_axis:
            # Nothing to do yet.
            return

        extent = ax.get_extent(lonlat=True)

        gone_home = False
        if np.all(np.isclose(ax.get_extent(lonlat=False), self._initial_extent_xy)):
            gone_home = True

            if self._reprojected:
                self._remove_change_axis_callbacks()
                ax.update_projection(self._crs_orig)
                self._initialize_axes(self._initial_extent_lonlat)
                self._add_change_axis_callbacks()

        # Reset to new extent
        self._changed_x_axis = False
        self._changed_y_axis = False
        self._extent_xy = extent_xy

        # This synchronizes the axis artist to the plot axes after zoom.
        if self._aa is not None:
            self._aa.set_position(self._ax.get_position(), which='original')

        self._draw_aa_bounds_and_labels()

        if gone_home:
            lon_range = self._redraw_dict['lon_range_home']
            lat_range = self._redraw_dict['lat_range_home']
        else:
            lon_range = [min(extent[0], extent[1]), max(extent[0], extent[1])]
            lat_range = [extent[2], extent[3]]

        if self._n_grid_lon is None or self._n_grid_lat is None:
            n_grid_lon, n_grid_lat = self._compute_n_grid_from_extent(
                [lon_range[0], lon_range[1], lat_range[0], lat_range[1]]
            )

            self._update_grid_locators(n_grid_lon, n_grid_lat)

        if self._redraw_dict['hpxmap'] is not None:
            lon_raster, lat_raster, values_raster = hpxmap_to_xy(self._redraw_dict['hpxmap'],
                                                                 lon_range,
                                                                 lat_range,
                                                                 nest=self._redraw_dict['nest'],
                                                                 xsize=self._redraw_dict['xsize'])
        elif self._redraw_dict['hspmap'] is not None:
            lon_raster, lat_raster, values_raster = hspmap_to_xy(self._redraw_dict['hspmap'],
                                                                 lon_range,
                                                                 lat_range,
                                                                 xsize=self._redraw_dict['xsize'])
        else:
            # Nothing to do
            return

        redraw_colorbar = False
        redraw_inset_colorbar = False
        if self._autorescale:
            # Recompute scaling
            try:
                vmin, vmax = get_autoscale_vmin_vmax(values_raster.compressed(), None, None)

                self._redraw_dict['vmin'] = vmin
                self._redraw_dict['vmax'] = vmax

                if self._redraw_dict['colorbar']:
                    redraw_colorbar = True
                if self._redraw_dict['inset_colorbar']:
                    redraw_inset_colorbar = True
            except IndexError:
                # We have zoomed to a blank spot, don't rescale
                vmin = self._redraw_dict['vmin']
                vmax = self._redraw_dict['vmax']
        else:
            vmin = self._redraw_dict['vmin']
            vmax = self._redraw_dict['vmax']

        if self._redraw_dict['im'] is not None:
            self._redraw_dict['im'].remove()
            self._redraw_dict['im'] = None

        im = self.pcolormesh(lon_raster, lat_raster, values_raster,
                             vmin=vmin, vmax=vmax,
                             rasterized=self._redraw_dict['rasterized'],
                             **self._redraw_dict['kwargs_pcolormesh'])
        self._redraw_dict['im'] = im
        self._ax._sci(im)

        if redraw_colorbar:
            mappable = ScalarMappable(Normalize(vmin=vmin, vmax=vmax),
                                      cmap=self._redraw_dict['colorbar'].cmap)
            self._redraw_dict['colorbar'].update_normal(mappable)
        if redraw_inset_colorbar:
            mappable = ScalarMappable(Normalize(vmin=vmin, vmax=vmax),
                                      cmap=self._redraw_dict['inset_colorbar'].cmap)
            self._redraw_dict['inset_colorbar'].update_normal(mappable)

    def _change_size(self, event):
        """Callback for figure resize.

        Parameters
        ----------
        event : `matplotlib.backend_bases.Event`
        """
        # This synchronizes the axis artist to the plot axes after zoom.
        if self._aa is not None:
            self._aa.set_position(self._ax.get_position(), which='original')

    def _draw_callback(self, event):
        # On draw it's sometimes necessary to synchronize the axisartist.
        if self._aa is not None:
            self._aa.set_position(self._ax.get_position(), which='original')

        # We need to set the initial extent on first draw
        if np.allclose(self._initial_extent_xy, 0):
            self._initial_extent_xy = self._ax.get_extent(lonlat=False)
            self._initial_extent_lonlat = self._ax.get_extent(lonlat=True)

    def _keypress_callback(self, event):
        if event.key == 'R':
            self._remove_change_axis_callbacks()
            extent = self._ax.get_extent(lonlat=True)

            lon_0_new = (extent[0] + extent[1])/2.
            if self.lat_0 is not None:
                lat_0_new = (extent[2] + extent[3])/2.
            else:
                lat_0_new = None

            # Decide if gnomonic or not
            if (extent[1] - extent[0])/2. < 1.0 and (extent[3] - extent[2])/2. < 1.0:
                # Make this a gnomonic projection
                crs_new = GnomonicCRS(lon_0=lon_0_new, lat_0=(extent[2] + extent[3])/2.)
            else:
                crs_new = self._crs_orig.with_new_center(lon_0_new, lat_0=lat_0_new)

            self._ax.update_projection(crs_new)
            self._initialize_axes(extent)
            self._changed_x_axis = True
            self._changed_y_axis = True
            self._change_axis(self._ax)
            self._add_change_axis_callbacks()
            plt.draw()

            self._reprojected = True

    def set_xlabel(self, text, side='bottom', **kwargs):
        """Set the label on the x axis.

        Parameters
        ----------
        text : `str`
            x label string.
        side : `str`, optional
            Side to set the label.  Can be ``bottom`` or ``top``.
        **kwargs : `dict`
            Additional keyword arguments accepted by ax.set_xlabel().
        """
        return self._aa.axis[side].label.set(text=text, **kwargs)

    def set_ylabel(self, text, side='left', **kwargs):
        """Set the label on the y axis.

        Parameters
        ----------
        text : `str`
            x label string.
        side : `str`, optional
            Side to set the label.  Can be ``left`` or ``right``.
        **kwargs : `dict`
            Additional keyword arguments accepted by ax.set_xlabel().
        """
        return self._aa.axis[side].label.set(text=text, **kwargs)

    @property
    def ax(self):
        return self._ax

    @property
    def aa(self):
        return self._aa

    def compute_extent(self, lon, lat):
        """Compute plotting extent for a set of lon/lat points.

        Uses a simple looping algorithm to find the ideal range so that
        all the points fit within the projected frame, with a small border.

        Parameters
        ----------
        lon : `np.ndarray`
            Array of longitude values.
        lat : `np.ndarray`
            Array of latitude values.

        Returns
        -------
        extent : `list`
            Plotting extent [lon_max, lon_min, lat_min, lat_max]
        """
        lon_wrap = wrap_values(lon)

        # Compute lat range with cushion
        lat_min0 = np.min(lat)
        lat_max0 = np.max(lat)
        lat_range = lat_max0 - lat_min0
        lat_min = np.clip(lat_min0 - 0.05*lat_range, -90.0, None)
        lat_max = np.clip(lat_max0 + 0.05*lat_range, None, 90.0)

        # Compute an ideally fitting lon range with cushion
        x, y = self.proj(lon, lat)

        lon_min0 = np.min(lon_wrap)
        lon_max0 = np.max(lon_wrap)
        lon_step = (lon_max0 - lon_min0)/20.
        lon_cent = (lon_min0 + lon_max0)/2.

        # Compute lon_min so that it fits all the data.
        enclosed = False
        lon_min = lon_cent - lon_step
        while not enclosed and lon_min > (self.lon_0 - 180.0):
            e_x, e_y = self.proj([lon_min, lon_min], [lat_min, lat_max])
            n_out = np.sum(x < e_x.min())
            if n_out == 0:
                enclosed = True
            else:
                lon_min = np.clip(lon_min - lon_step, self.lon_0 - 180., None)

        # Compute lon_max so that it fits all the data
        enclosed = False
        lon_max = lon_cent + lon_step
        while not enclosed and lon_max < (self.lon_0 + 180.0):
            e_x, e_y = self.proj([lon_max, lon_max], [lat_min, lat_max])
            n_out = np.sum(x > e_x.max())
            if n_out == 0:
                enclosed = True
            else:
                lon_max = np.clip(lon_max + lon_step, None, self.lon_0 + 180.)

        return [lon_max, lon_min, lat_min, lat_max]

    def plot(self, *args, **kwargs):
        return self._ax.plot(*args, **kwargs)

    def scatter(self, *args, **kwargs):
        return self._ax.scatter(*args, **kwargs)

    def pcolormesh(self, *args, **kwargs):
        return self._ax.pcolormesh(*args, **kwargs)

    def fill(self, *args, **kwargs):
        return self._ax.fill(*args, **kwargs)

    def circle(self, *args, **kwargs):
        return self._ax.circle(*args, **kwargs)

    def ellipse(self, *args, **kwargs):
        return self._ax.ellipse(*args, **kwargs)

    def legend(self, *args, loc='upper left', **kwargs):
        """Add legend to the axis with ax.legend(*args, **kwargs)."""
        return self._ax.legend(*args, loc=loc, **kwargs)

    def draw_polygon(self, lon, lat, edgecolor='red', linestyle='solid',
                     facecolor=None, **kwargs):
        """Plot a polygon from a list of lon, lat coordinates.

        This routine is a convenience wrapper around plot() and fill(), both
        of which work in geodesic coordinates.

        Parameters
        ----------
        lon : `np.ndarray`
            Array of longitude points in polygon.
        lat : `np.ndarray`
            Array of latitude points in polygon.
        edgecolor : `str`, optional
            Color of polygon boundary.  Set to None for no boundary.
        linestyle : `str`, optional
            Line style for boundary.
        facecolor : `str`, optional
            Color of polygon face.  Set to None for no fill color.
        **kwargs : `dict`, optional
            Additional keywords passed to plot.
        """
        if linestyle is not None and edgecolor is not None:
            self.plot(np.append(lon, lon[0]),
                      np.append(lat, lat[0]),
                      color=edgecolor, linestyle=linestyle, **kwargs)
        if facecolor is not None:
            self.fill(lon, lat, color=facecolor, **kwargs)

    def draw_polygon_file(self, filename, reverse=True,
                          edgecolor='red', linestyle='solid', **kwargs):
        """Draw a text file containing lon, lat coordinates of polygon(s).

        Parameters
        ----------
        filename : `str`
            Name of file containing the polygon(s) [lon, lat, poly]
        reverse : `bool`
            Reverse drawing order of points in each polygon.
        edgecolor : `str`
            Color of polygon boundary.
        linestyle : `str`, optional
            Line style for boundary.
        **kwargs : `dict`
            Additional keywords passed to plot.
        """
        try:
            data = np.genfromtxt(filename, names=['lon', 'lat', 'poly'])
        except ValueError:
            from numpy.lib.recfunctions import append_fields
            data = np.genfromtxt(filename, names=['lon', 'lat'])
            data = append_fields(data, 'poly', np.zeros(len(data)))

        for p in np.unique(data['poly']):
            poly = data[data['poly'] == p]
            lon = poly['lon'][::-1] if reverse else poly['lon']
            lat = poly['lat'][::-1] if reverse else poly['lat']
            self.draw_polygon(lon,
                              lat,
                              edgecolor=edgecolor,
                              linestyle=linestyle,
                              **kwargs)
            # Only add the label to the first polygon plotting.
            kwargs.pop('label', None)

    @add_func_docstr(draw_hpxmap_docstr)
    def draw_hpxmap(self, hpxmap, nest=False, zoom=True, xsize=1000, vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):

        nside = hpg.npixel_to_nside(hpxmap.size)
        pixels, = np.where(hpxmap != hpg.UNSEEN)

        if lon_range is None or lat_range is None:
            if zoom:
                _lon_range, _lat_range = healpix_pixels_range(nside,
                                                              pixels,
                                                              self._wrap,
                                                              nest=nest)
            else:
                extent = self.get_extent()
                lon_range = [min(extent[0], extent[1]), max(extent[0], extent[1])]
                lat_range = [extent[2], extent[3]]

            if lon_range is None:
                lon_range = _lon_range
            if lat_range is None:
                lat_range = _lat_range

        # FIXME: add aspect ratio
        lon_raster, lat_raster, values_raster = hpxmap_to_xy(hpxmap,
                                                             lon_range,
                                                             lat_range,
                                                             nest=nest,
                                                             xsize=xsize)

        if vmin is None or vmax is None:
            vmin, vmax = get_autoscale_vmin_vmax(values_raster.compressed(), vmin, vmax)

        if zoom:
            extent = self.compute_extent(lon_raster[:-1, :-1][~values_raster.mask],
                                         lat_raster[:-1, :-1][~values_raster.mask])
            self.set_extent(extent)

        im = self.pcolormesh(
            lon_raster,
            lat_raster,
            values_raster,
            vmin=vmin,
            vmax=vmax,
            rasterized=rasterized,
            **kwargs,
        )
        self._ax._sci(im)

        # Link up callbacks
        self._redraw_dict['hspmap'] = None
        self._redraw_dict['hpxmap'] = hpxmap
        self._redraw_dict['lon_range_home'] = lon_range
        self._redraw_dict['lat_range_home'] = lat_range
        self._redraw_dict['nside'] = nside
        self._redraw_dict['nest'] = nest
        self._redraw_dict['vmin'] = vmin
        self._redraw_dict['vmax'] = vmax
        self._redraw_dict['xsize'] = xsize
        self._redraw_dict['rasterized'] = rasterized
        self._redraw_dict['kwargs_pcolormesh'] = kwargs

        return im, lon_raster, lat_raster, values_raster

    @add_func_docstr(draw_hpxpix_docstr)
    def draw_hpxpix(self, nside, pixels, values, nest=False, zoom=True, xsize=1000,
                    vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        if lon_range is None or lat_range is None:
            if zoom:
                _lon_range, _lat_range = healpix_pixels_range(nside,
                                                              pixels,
                                                              self._wrap,
                                                              nest=nest)
            else:
                extent = self.get_extent()
                lon_range = [min(extent[0], extent[1]), max(extent[0], extent[1])]
                lat_range = [extent[2], extent[3]]

            if lon_range is None:
                lon_range = _lon_range
            if lat_range is None:
                lat_range = _lat_range

        # FIXME: add aspect ratio
        lon_raster, lat_raster, values_raster = healpix_to_xy(
            nside,
            pixels,
            values,
            nest=nest,
            xsize=xsize,
            lon_range=lon_range,
            lat_range=lat_range
        )

        if vmin is None or vmax is None:
            vmin, vmax = get_autoscale_vmin_vmax(values_raster.compressed(), vmin, vmax)

        if zoom:
            extent = self.compute_extent(lon_raster[:-1, :-1][~values_raster.mask],
                                         lat_raster[:-1, :-1][~values_raster.mask])
            self.set_extent(extent)

        im = self.pcolormesh(
            lon_raster,
            lat_raster,
            values_raster,
            vmin=vmin,
            vmax=vmax,
            rasterized=rasterized,
            **kwargs,
        )
        self._ax._sci(im)

        return im, lon_raster, lat_raster, values_raster

    @add_func_docstr(draw_hspmap_docstr)
    def draw_hspmap(self, hspmap, zoom=True, xsize=1000, vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, valid_mask=False,
                    **kwargs):
        self._hspmap = hspmap
        self._hpxmap = None

        if lon_range is None or lat_range is None:
            if zoom:
                # Using the coverage map is much faster even if approximate.
                _lon_range, _lat_range = healpix_pixels_range(hspmap.nside_coverage,
                                                              np.where(hspmap.coverage_mask)[0],
                                                              self._wrap,
                                                              nest=True)
            else:
                extent = self.get_extent()
                lon_range = [min(extent[0], extent[1]), max(extent[0], extent[1])]
                lat_range = [extent[2], extent[3]]

            if lon_range is None:
                lon_range = _lon_range
            if lat_range is None:
                lat_range = _lat_range

        if hspmap.is_rec_array and not valid_mask:
            warnings.warn(
                """
                draw_hspmap called with a record array HealSparseMap.  Assuming valid_mask=True.
                To instead visualize component "A" of the record array, draw_hspmap(hspmap["A"])
                """
            )
            valid_mask = True

        # FIXME: add aspect ratio
        lon_raster, lat_raster, values_raster = hspmap_to_xy(hspmap,
                                                             lon_range,
                                                             lat_range,
                                                             xsize=xsize,
                                                             valid_mask=valid_mask)

        if vmin is None or vmax is None:
            vmin, vmax = get_autoscale_vmin_vmax(values_raster.compressed(), vmin, vmax)

        if zoom:
            # Watch for masked array here...
            extent = self.compute_extent(lon_raster[:-1, :-1][~values_raster.mask],
                                         lat_raster[:-1, :-1][~values_raster.mask])
            self.set_extent(extent)

        im = self.pcolormesh(
            lon_raster,
            lat_raster,
            values_raster,
            vmin=vmin,
            vmax=vmax,
            rasterized=rasterized,
            **kwargs,
        )

        self._ax._sci(im)

        # Link up callbacks
        self._redraw_dict['hspmap'] = hspmap
        self._redraw_dict['hpxmap'] = None
        self._redraw_dict['lon_range_home'] = lon_range
        self._redraw_dict['lat_range_home'] = lat_range
        self._redraw_dict['im'] = im
        self._redraw_dict['vmin'] = vmin
        self._redraw_dict['vmax'] = vmax
        self._redraw_dict['xsize'] = xsize
        self._redraw_dict['rasterized'] = rasterized
        self._redraw_dict['kwargs_pcolormesh'] = kwargs

        return im, lon_raster, lat_raster, values_raster

    @add_func_docstr(draw_hpxbin_docstr)
    def draw_hpxbin(self, lon, lat, C=None, nside=256, nest=False, zoom=True, xsize=1000,
                    vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        hpxmap = healpix_bin(lon, lat, C=C, nside=nside, nest=nest)

        im, lon_raster, lat_raster, values_raster = self.draw_hpxmap(
            hpxmap, nest=nest, zoom=zoom, xsize=xsize, vmin=vmin,
            vmax=vmax, rasterized=rasterized, lon_range=lon_range,
            lat_range=lat_range,
            **kwargs)

        return hpxmap, im, lon_raster, lat_raster, values_raster

    def draw_inset_colorbar(self, format=None, label=None, ticks=None, fontsize=11,
                            width="25%", height="5%", loc=7, bbox_to_anchor=(0., -0.04, 1, 1),
                            orientation='horizontal', ax=None, **kwargs):
        """Draw an inset colorbar.

        Parameters
        ----------
        format : `str`, optional
            Format string for tick labels.
        label : `str`, optional
            Label to attach to inset colorbar.
        ticks : `list`, optional
            List of tick values.
        fontsize : `int`, optional
            Font size to use for ticks.
        width : `str`, optional
            Fraction of total axis width for inset colorbar.
        height : `str`, optional
            Fraction of total axis height for inset colorbar.
        loc : `int`, optional
            Matplotlib location code.
        bbox_to_anchor : `tuple`, optional
            Where to put inset colorbar bbox.
        orientation : `str`, optional
            Inset colorbar orientation (``horizontal`` or ``vertical``).
        ax : `SkyAxesSubplot`, optional
            Axis associated with inset colorbar.  If None, use
            skyaxes associated with map.
        **kwargs : `dict`, optional
            Additional kwargs to pass to inset_axes or colorbar.

        Returns
        -------
        colorbar : `matplotlib.colorbar.Colorbar`
        colorbar_axis : `mpl_toolkits.axes_grid1.parasite_axes.AxesHostAxes`
        """
        if ax is None:
            ax = self._ax
        im = plt.gci()
        cax = inset_axes(ax,
                         width=width,
                         height=height,
                         loc=loc,
                         bbox_to_anchor=bbox_to_anchor,
                         bbox_transform=ax.transAxes,
                         **kwargs)
        cmin, cmax = im.get_clim()

        if ticks is None and cmin is not None and cmax is not None:
            cmed = (cmax + cmin)/2.
            delta = (cmax - cmin)/10.
            ticks = np.array([cmin + delta, cmed, cmax - delta])

        tmin = np.min(np.abs(ticks[0]))
        tmax = np.max(np.abs(ticks[1]))

        if format is None:
            if (tmin < 1e-2) or (tmax > 1e3):
                format = '$%.1e$'
            elif (tmin > 0.1) and (tmax < 100):
                format = '$%.1f$'
            elif (tmax > 100):
                format = '$%i$'
            else:
                format = '$%.2g$'

        custom_format = False
        if format == 'custom':
            custom_format = True
            ticks = np.array([cmin, 0.85*cmax])
            format = '$%.0e$'

        cbar = plt.colorbar(cax=cax, orientation=orientation, ticks=ticks, format=format, **kwargs)
        cax.xaxis.set_ticks_position('top')
        cax.tick_params(axis='x', labelsize=fontsize)

        if custom_format:
            ticklabels = cax.get_xticklabels()
            for i, lab in enumerate(ticklabels):
                val, exp = ticklabels[i].get_text().split('e')
                ticklabels[i].set_text(r'$%s \times 10^{%i}$'%(val, int(exp)))
            cax.set_xticklabels(ticklabels)

        if label is not None:
            cbar.set_label(label, size=fontsize)
            cax.xaxis.set_label_position('top')

        plt.sca(ax)

        # Save reference to colorbar for zooming
        cbar_kwargs = kwargs
        cbar_kwargs['format'] = format
        cbar_kwargs['label'] = label
        cbar_kwargs['ticks'] = ticks
        cbar_kwargs['fontsize'] = fontsize
        cbar_kwargs['width'] = width
        cbar_kwargs['height'] = height
        cbar_kwargs['loc'] = loc
        cbar_kwargs['bbox_to_anchor'] = bbox_to_anchor
        cbar_kwargs['orientation'] = orientation
        self._redraw_dict['inset_colorbar'] = cbar
        self._redraw_dict['inset_colorbar_kwargs'] = cbar_kwargs

        return cbar, cax

    def draw_colorbar(self, label=None, ticks=None, fontsize=11,
                      fraction=0.15, location='right', pad=0.0,
                      ax=None, **kwargs):
        """Draw a colorbar.

        Parameters
        ----------
        label : `str`, optional
            Label to attach to colorbar.
        ticks : `list`, optional
            List of tick values.
        fontsize : `int`, optional
            Font size to use for ticks.
        fraction : `float`, optional
            Fraction of original axes to use for colorbar.
        location : `str`, optional
            Colorbar location (``right``, ``bottom``, ``left``, ``top``).
        pad : `float`, optional
            Fraction of original axes between colorbar and original axes.
        ax : `SkyAxesSubplot`, optional
            Axis associated with inset colorbar.  If None, use
            skyaxes associated with map.
        **kwargs : `dict`, optional
            Additional kwargs to send to colorbar().

        Returns
        -------
        colorbar : `matplotlib.colorbar.Colorbar`
        """
        if ax is None:
            ax = self._ax

        cbar = plt.colorbar(ax=ax, location=location, ticks=ticks,
                            fraction=fraction, pad=pad, **kwargs)

        if location == 'right' or location == 'left':
            cbar_axis = 'y'
        else:
            cbar_axis = 'x'

        cbar.ax.tick_params(axis=cbar_axis, labelsize=fontsize)

        if label is not None:
            cbar.set_label(label, size=fontsize)

        if self._aa is not None:
            self._aa.set_position(self._ax.get_position(), which='original')

        # Reset the "home" position because axis has been shifted.
        self._initial_extent_xy = self._ax.get_extent(lonlat=False)

        plt.sca(ax)

        # Save reference to colorbar for zooming
        cbar_kwargs = kwargs
        cbar_kwargs['label'] = label
        cbar_kwargs['ticks'] = ticks
        cbar_kwargs['fontsize'] = fontsize
        cbar_kwargs['fraction'] = fraction
        cbar_kwargs['location'] = location
        cbar_kwargs['pad'] = pad
        self._redraw_dict['colorbar'] = cbar
        self._redraw_dict['colorbar_kwargs'] = cbar_kwargs

        return cbar

    def draw_milky_way(self, width=10, linewidth=1.5, color='black', linestyle='-', **kwargs):
        """Draw the Milky Way galaxy.

        Parameters
        ----------
        width : `float`
            Number of degrees north and south to draw dotted lines.
        linewidth : `float`
            Width of line along the plane.
        color : `str`
            Color of Milky Way plane.
        linestyle : `str`
            Style of line.
        **kwargs : `dict`
            Additional kwargs to pass to plot.
        """
        from astropy import units as u
        from astropy.coordinates import SkyCoord

        glon = np.linspace(0, 360, 500)
        glat = np.zeros_like(glon)

        if not self._galactic:
            gc = SkyCoord(l=glon*u.degree, b=glat*u.degree, frame='galactic')
            radec = gc.fk5
            lon = radec.ra.degree
            lat = radec.dec.degree
        else:
            lon = glon
            lat = glat

        self.plot(lon, lat, linewidth=linewidth, color=color, linestyle=linestyle, **kwargs)
        # pop any labels
        kwargs.pop('label', None)
        if width > 0:
            for delta in [+width, -width]:
                if not self._galactic:
                    gc = SkyCoord(l=glon*u.degree, b=(glat + delta)*u.degree, frame='galactic')
                    radec = gc.fk5
                    lon = radec.ra.degree
                    lat = radec.dec.degree
                else:
                    lon = glon
                    lat = glat + delta
                self.plot(lon, lat, linewidth=1.0, color=color,
                          linestyle='--', **kwargs)

    def tissot_indicatrices(self, radius=5.0, num_lon=9, num_lat=5, color='red', alpha=0.5):
        """Draw Tissot indicatrices.

        See https://en.wikipedia.org/wiki/Tissot%27s_indicatrix for details.

        Parameters
        ----------
        radius : `float`
            Radius of each indicatrix circle.
        num_lon : `int`
            Number of indicatrices in the longitude direction.
        num_lat : `int`
            Number of indicatrices in the latitude direction.
        color : `str`, optional
            Color of indicatrices.
        alpha : `float`, optional
            Alpha of indicatrices.
        """
        lons = np.linspace(-175.0, 175.0, num_lon)
        lats = np.linspace(-80.0, 80.0, num_lat)

        for lat in lats:
            # We want to skip alternate indicatrices at high latitudes.
            skip_alternate = False
            if np.abs(lat) >= 75.0:
                skip_alternate = True
            skipped = False
            for lon in lons:
                if skip_alternate and not skipped:
                    skipped = True
                    continue
                _ = self._ax.circle(lon, lat, radius, fill=True, color=color, alpha=alpha)
                skipped = False

    @property
    def lon_0(self):
        return self._ax.lon_0

    @property
    def lat_0(self):
        return self._ax.lat_0

    @property
    def crs(self):
        return self._ax.projection

    @property
    def projection_name(self):
        return self._ax.projection.name

    @property
    def _full_sky_extent_initial(self):
        return [self.lon_0 - 180.0,
                self.lon_0 + 180.0,
                -90.0 + self._pole_clip,
                90.0 - self._pole_clip]

    @property
    def _pole_clip(self):
        # Allow clipping of the poles until Mollweide is fixed in proj
        return 0.0

    @property
    def _full_circle(self):
        # Is this projection a full circle?
        return False

    @property
    def _equatorial_labels(self):
        # Should the longitude labels be along the equator?
        return False

    @property
    def _radial_labels(self):
        # Are there radial labels?
        return False

    @property
    def _init_extent_xy(self):
        # Is the initial extent in x/y space?
        return False

    @property
    def _default_xy_labels(self):
        # Default labels in x, y
        if self._galactic:
            return ("Galactic Longitude", "Galactic Latitude")
        else:
            return ("Right Ascension", "Declination")



import numpy as np

from ._skyproj import _Skyproj

from ._docstrings import skyproj_init_parameters, skyproj_kwargs_par


__all__ = ['Skyproj', 'McBrydeSkyproj', 'LaeaSkyproj', 'MollweideSkyproj',
           'HammerSkyproj', 'EqualEarthSkyproj', 'GnomonicSkyproj',
           'ObliqueMollweideSkyproj', 'AlbersSkyproj']


class _Stadium:
    """Extension class to create a stadium-shaped projection boundary.
    """
    def _compute_proj_boundary_xy(self):
        proj_boundary_xy = {}

        edge_offset = 180.0 - 1e-6
        nstep = 1000

        x, y = self.proj(np.linspace(self.lon_0 - edge_offset,
                                     self.lon_0 - edge_offset,
                                     nstep),
                         np.linspace(-90.0 + 1e-6,
                                     90.0 - 1e-6,
                                     nstep))
        proj_boundary_xy['left'] = np.column_stack((x, y))

        x, y = self.proj(np.linspace(self.lon_0 + edge_offset,
                                     self.lon_0 + edge_offset,
                                     nstep),
                         np.linspace(-90.0 + 1e-6,
                                     90.0 - 1e-6,
                                     nstep))
        proj_boundary_xy['right'] = np.column_stack((x, y))

        x, y = self.proj(np.linspace(self.lon_0 - edge_offset,
                                     self.lon_0 + edge_offset,
                                     nstep),
                         np.linspace(90.0 - 1e-6,
                                     90.0 - 1e-6,
                                     nstep))
        proj_boundary_xy['top'] = np.column_stack((x, y))

        x, y = self.proj(np.linspace(self.lon_0 - edge_offset,
                                     self.lon_0 + edge_offset,
                                     nstep),
                         np.linspace(-90.0 + 1e-6,
                                     -90.0 + 1e-6,
                                     nstep))
        proj_boundary_xy['bottom'] = np.column_stack((x, y))

        return proj_boundary_xy


class _Ellipse21:
    """Extension class to create an ellipse-shaped projection boundary.
    """
    def _compute_proj_boundary_xy(self):
        proj_boundary_xy = {}

        nstep = 1000

        t = np.linspace(-np.pi/2., np.pi/2., nstep)
        x = 2*self.crs.radius*np.sqrt(2)*np.cos(t)
        y = self.crs.radius*np.sqrt(2)*np.sin(t)
        proj_boundary_xy['right'] = np.column_stack((x, y))

        t = np.linspace(np.pi/2., 3*np.pi/2., nstep)
        x = 2*self.crs.radius*np.sqrt(2)*np.cos(t)
        y = self.crs.radius*np.sqrt(2)*np.sin(t)
        proj_boundary_xy['left'] = np.column_stack((x, y))

        proj_boundary_xy['top'] = np.zeros((0, 2))
        proj_boundary_xy['bottom'] = np.zeros((0, 2))

        return proj_boundary_xy


class _Circle:
    """Extension class to create a circular projection boundary.
    """
    def _compute_proj_boundary_xy(self):
        proj_boundary_xy = {}

        nstep = 1000

        t = np.linspace(-np.pi/2., np.pi/2., nstep)
        x = 2*self.crs.radius*np.cos(t)
        y = 2*self.crs.radius*np.sin(t)
        proj_boundary_xy['right'] = np.column_stack((x, y))

        t = np.linspace(np.pi/2., 3*np.pi/2., nstep)
        x = 2*self.crs.radius*np.cos(t)
        y = 2*self.crs.radius*np.sin(t)
        proj_boundary_xy['left'] = np.column_stack((x, y))

        proj_boundary_xy['top'] = np.zeros((0, 2))
        proj_boundary_xy['bottom'] = np.zeros((0, 2))

        return proj_boundary_xy


# The default skyproj is a cylindrical Plate Carree projection.

class Skyproj(_Skyproj, _Stadium):
    __doc__ = (skyproj_init_parameters("A Plate Carree cylindrical projection Skyproj map.")
               + skyproj_kwargs_par)

    # Plate Carree
    def __init__(
        self,
        ax=None,
        *,
        lon_0=0.0,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='cyl',
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )


# The following skyprojs include the equal-area projections that are tested
# and known to work.

class McBrydeSkyproj(_Skyproj, _Stadium):
    __doc__ = (skyproj_init_parameters("A McBryde-Thomas Flat Polar Quartic projection Skyproj map.")
               + skyproj_kwargs_par)

    # McBryde-Thomas Flat Polar Quartic
    def __init__(
        self,
        ax=None,
        *,
        lon_0=0.0,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='mbtfpq',
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )


class LaeaSkyproj(_Skyproj, _Circle):
    __doc__ = (skyproj_init_parameters("A Lambert Azimuthal Equal Area projection Skyproj map.")
               + skyproj_kwargs_par)
    __doc__ = skyproj_init_parameters(
        "A Lambert Azimuthal Equal Area projection Skyproj map.",
        include_lon_0=False,
    )
    __doc__ += """
lon_0 : `float`, optional
    Central longitude of the LAEA projection.
lat_0 : `float`, optional
    Central latitude of the LAEA projection."""

    # Lambert Azimuthal Equal Area
    def __init__(
        self,
        ax=None,
        *,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        lon_0=0.0,
        lat_0=0.0,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='laea',
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            lon_0=lon_0,
            lat_0=lat_0,
            **kwargs,
        )

    @property
    def _full_circle(self):
        return True

    @property
    def _init_extent_xy(self):
        return True

    @property
    def _radial_labels(self):
        return True

    @property
    def _default_xy_labels(self):
        return ("", "")

    @property
    def _full_sky_extent_initial(self):
        lon0 = self.lon_0 - 180.0
        lon1 = self.lon_0 + 180.0
        _lat_0 = self.lat_0
        if _lat_0 == -90.0:
            lat0 = -90.0
            lat1 = 90.0 - 1e-5
        elif _lat_0 == 90.0:
            lat0 = -90.0 + 1e-5
            lat1 = 90.0
        else:
            lat0 = -90.0 + 1e-5
            lat1 = 90.0 - 1e-5

        return [lon0, lon1, lat0, lat1]


class MollweideSkyproj(_Skyproj, _Ellipse21):
    __doc__ = (skyproj_init_parameters("A Mollweide projection Skyproj map.")
               + skyproj_kwargs_par)

    # Mollweide
    def __init__(
        self,
        ax=None,
        *,
        lon_0=0.0,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='moll',
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )

    @property
    def _pole_clip(self):
        return 1.0

    @property
    def _equatorial_labels(self):
        return True


class HammerSkyproj(_Skyproj, _Ellipse21):
    __doc__ = (skyproj_init_parameters("A Hammer-Aitoff projection Skyproj map.")
               + skyproj_kwargs_par)

    # Hammer-Aitoff
    def __init__(
        self,
        ax=None,
        *,
        lon_0=0.0,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='hammer',
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )

    @property
    def _equatorial_labels(self):
        return True


class EqualEarthSkyproj(_Skyproj, _Stadium):
    __doc__ = (skyproj_init_parameters("An Equal Earth projection Skyproj map.")
               + skyproj_kwargs_par)

    # Equal Earth
    def __init__(
        self,
        ax=None,
        *,
        lon_0=0.0,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='eqearth',
            lon_0=lon_0,
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            **kwargs,
        )


class ObliqueMollweideSkyproj(_Skyproj, _Ellipse21):
    __doc__ = skyproj_init_parameters("An Oblique Mollweide projection Skyproj map.", include_lon_0=False)
    __doc__ += """
lon_0 : `float`, optional
    Central longitude of the underlying Mollweide projection.
lat_p : `float`, optional
    Latitude of the North Pole of the unrotated coordinate system.
lon_p : `float`, optional
    Longitude of the North Pole of the unrotated coordinate system."""
    __doc__ += skyproj_kwargs_par

    # Oblique Mollweide
    def __init__(
        self,
        ax=None,
        *,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        lon_0=0.0,
        lat_p=90.0,
        lon_p=0.0,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='obmoll',
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            lon_0=lon_0,
            lat_p=lat_p,
            lon_p=lon_p,
            **kwargs,
        )

    @property
    def _pole_clip(self):
        return 1.0

    @property
    def _equatorial_labels(self):
        return True

    @property
    def _default_xy_labels(self):
        return ("", "")

    @property
    def _init_extent_xy(self):
        return True


# The Gnomonic (tangent plane) projection is not equal-area and
# is not available for full-sky plots.  It is only for small
# zoomed regions
class GnomonicSkyproj(_Skyproj, _Circle):
    __doc__ = skyproj_init_parameters("""A Gnomonic (tangent plane) projection Skyproj map.

This projection is not equal area and is not available for full sky plots.
It should only be used for small zoomed regions.""", include_lon_0=False)
    __doc__ += """
lon_0 : `float`, optional
    Central longitude of the Gnomonic projection.
lat_0 : `float`, optional
    Central latitude of the Gnomonic projection."""
    __doc__ += skyproj_kwargs_par

    # Gnomonic
    def __init__(
        self,
        ax=None,
        *,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        lon_0=0,
        lat_0=0,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='gnom',
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            lon_0=lon_0,
            lat_0=lat_0,
            **kwargs,
        )

    @property
    def _full_sky_extent_initial(self):
        lon_0 = self.lon_0
        lat_0 = self.lat_0
        cos_lat = np.cos(np.deg2rad(lat_0))
        return [lon_0 - 0.5/cos_lat, lon_0 + 0.5/cos_lat,
                lat_0 - 0.5, lat_0 + 0.5]


class AlbersSkyproj(_Skyproj, _Stadium):
    __doc__ = skyproj_init_parameters("An Albers Equal Area Skyproj map.", include_lon_0=False)
    __doc__ += """
lon_0 : `float`, optional
    Central longitude of the projection.
lat_1 : `float`, optional
    First standard parallel of the projection.
lat_2 : `float`, optional
    Second standard parallel of the projection."""
    __doc__ += skyproj_kwargs_par

    # Albers Equal Area
    def __init__(
        self,
        ax=None,
        *,
        gridlines=True,
        celestial=True,
        extent=None,
        longitude_ticks='positive',
        autorescale=True,
        galactic=False,
        rcparams={},
        n_grid_lon=None,
        n_grid_lat=None,
        min_lon_ticklabel_delta=0.1,
        lon_0=0,
        lat_1=15.0,
        lat_2=45.0,
        **kwargs,
    ):
        super().__init__(
            ax=ax,
            projection_name='aea',
            gridlines=gridlines,
            celestial=celestial,
            extent=extent,
            longitude_ticks=longitude_ticks,
            autorescale=autorescale,
            galactic=galactic,
            rcparams=rcparams,
            n_grid_lon=n_grid_lon,
            n_grid_lat=n_grid_lat,
            min_lon_ticklabel_delta=min_lon_ticklabel_delta,
            lon_0=lon_0,
            lat_1=lat_1,
            lat_2=lat_2,
            **kwargs,
        )

    @property
    def _default_xy_labels(self):
        return ("", "")



import numpy as np

import skyproj


def test_wrap_values():
    """Test the wrap values.
    """
    # Make a giant array that wraps around 0 to 360 many times.
    arr = np.arange(20000.) - 10000.
    sin_arr = np.sin(np.deg2rad(arr))
    cos_arr = np.cos(np.deg2rad(arr))

    for wrap_angle in np.arange(360.):
        wrapped_arr = skyproj.utils.wrap_values(arr, wrap=wrap_angle)

        # Confirm we have the correct range:
        np.testing.assert_almost_equal(np.min(wrapped_arr), wrap_angle - 360.)
        np.testing.assert_almost_equal(np.max(wrapped_arr), wrap_angle - 1.0)

        np.testing.assert_array_almost_equal(np.sin(np.deg2rad(wrapped_arr)), sin_arr)
        np.testing.assert_array_almost_equal(np.cos(np.deg2rad(wrapped_arr)), cos_arr)



"""
Test notebooks in tutorial.

Adapted from:
https://blog.thedataincubator.com/2016/06/testing-jupyter-notebooks/
"""
import os
import pytest
import subprocess
import tempfile

try:
    import nbformat
except ImportError:
    nbformat = None


ROOT = os.path.abspath(os.path.dirname(__file__))


def _notebook_run(nbfile):
    """Execute a notebook via nbconvert and collect output.

    Parameters
    ----------
    nbfile : `str`
        Notebook file to run.

    Returns
    -------
    nb : `???`
        Parsed notebook object.
    errors : `list` [`str`]
        List of error strings.
    """
    with tempfile.NamedTemporaryFile(suffix=".ipynb") as fout:
        args = ["jupyter", "nbconvert", "--to", "notebook", "--execute",
                "--ExecutePreprocessor.timeout=60", "--log-level=WARN",
                "--output", fout.name, nbfile]
        subprocess.check_call(args)

        fout.seek(0)
        nb = nbformat.read(fout, nbformat.current_nbformat)

    errors = [output for cell in nb.cells if "outputs" in cell
              for output in cell["outputs"]
              if output.output_type == "error"]

    return nb, errors


@pytest.mark.skipif(nbformat is None, reason="nbformat not installed.")
@pytest.mark.parametrize("nbfile", ["tutorial_baseclass.ipynb",
                                    "tutorial_surveys.ipynb",
                                    "tutorial_healsparse.ipynb"])
def test_tutorial_notebooks(nbfile):
    """Test running a tutorial notebook."""
    fname = os.path.abspath(os.path.join(ROOT, '../tutorial', nbfile))
    nb, errors = _notebook_run(fname)
    assert errors == []



import os
import pytest

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize("skyproj", [skyproj.Skyproj,
                                     skyproj.McBrydeSkyproj])
def test_tissot(tmp_path, skyproj):
    """Test Tissot Indicatrices."""
    plt.rcParams.update(plt.rcParamsDefault)

    # Full image
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax)
    sp.tissot_indicatrices()
    fname = f'{sp.projection_name}_tissot.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)



import os

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


def test_skyproj_plotting(tmp_path):
    """Test basic plotting routines."""
    plt.rcParams.update(plt.rcParamsDefault)

    # Tests are consolidated for fewer comparisons.
    # Note that tests of wrapping lines and polygons are in
    # test_lines_polygons.
    # Tests of pcolormesh are in the healsparse/healpix map plots.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax, extent=[0, 50, 0, 50])

    # Test ``plot`` with points.
    sp.plot([10, 20, 30, 40], [10, 20, 30, 40], 'k+')
    sp.plot([40, 30, 20, 10], [10, 20, 30, 40], 'r.')

    # Test ``plot`` with lines.
    # Note that the geodesic line segments do not meet the interior
    # points plotted above.
    sp.plot([10, 40], [10, 40], 'k-')
    sp.plot([40, 10], [10, 40], 'r:')

    # Test ``fill``.
    sp.fill([20, 25, 25, 20], [20, 20, 25, 25], color='blue')

    # Test ``scatter`` with points.
    sp.scatter([15, 35], [15, 35], c=['magenta', 'orange'])

    fname = 'plotting_routines.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)



import os

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


def test_draw_milky_way(tmp_path):
    """Test drawing the Milky Way."""
    plt.rcParams.update(plt.rcParamsDefault)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    sp.draw_milky_way(label='Milky Way')
    sp.legend()
    fname = 'milky_way.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


def test_draw_milky_way_galactic(tmp_path):
    """Test drawing the Milky Way (Galactic Coordinates)."""
    plt.rcParams.update(plt.rcParamsDefault)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax, galactic=True, longitude_ticks='symmetric')
    sp.draw_milky_way(label='Milky Way')
    sp.legend()
    fname = 'milky_way_galactic.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)




"""
Random utilities
"""
import os
import os.path
import numpy as np

__all__ = ["get_datadir", "get_datafile", "wrap_values"]


AUTOSCALE_PERCENTILE_LOW = 2.5
AUTOSCALE_PERCENTILE_HIGH = 97.5


def get_datadir():
    from os.path import abspath, dirname, join
    return join(dirname(abspath(__file__)), 'data')


def get_datafile(filename):
    return os.path.join(get_datadir(), filename)


def wrap_values(values, wrap=180.0):
    """Wrap values according to the wrap angle.

    Parameters
    ----------
    values : `float` or `np.ndarray`
        Values to remap.
    wrap : `float`, optional
        Wrap angle to apply.

    Returns
    -------
    wrapped_array
    """
    return (values - wrap) % 360. + (wrap - 360.)


def _get_boundary_poly_xy(bounds_xy, extent_xy, proj, proj_inverse):
    """Get the boundary polygon in x/y space.

    Parameters
    ----------
    bounds_xy : `dict`
        Dictionary with projection boundaries.
    extent_xy : `list`
        Extent of bounding box.
    proj : `func`
        Forward projection function.
    proj_inverse : `func`
        Inverse projection function.

    Returns
    -------
    boundary_poly_xy : `np.ndarray`
        N,2 array of the boundary polygon.
    """
    bounds_xy_clipped_sides = []
    nstep = 1000

    if extent_xy[0] < extent_xy[1]:
        x0_index = 0
        x1_index = 1
    else:
        x0_index = 1
        x1_index = 0

    def _generate_side(side):
        if bounds_xy[side] is not None:
            use_bounds, = np.where((bounds_xy[side][:, 0] >= extent_xy[x0_index])
                                   & (bounds_xy[side][:, 0] <= extent_xy[x1_index])
                                   & (bounds_xy[side][:, 1] >= extent_xy[2])
                                   & (bounds_xy[side][:, 1] <= extent_xy[3]))
            bounds_x = bounds_xy[side][use_bounds, 0]
            bounds_y = bounds_xy[side][use_bounds, 1]
        else:
            bounds_x = []
            bounds_y = []

        if side == 'left':
            box_x = np.linspace(extent_xy[x0_index], extent_xy[x0_index], nstep)
            box_y = np.linspace(extent_xy[2], extent_xy[3], nstep)
        elif side == 'top':
            box_x = np.linspace(extent_xy[x0_index], extent_xy[x1_index], nstep)
            box_y = np.linspace(extent_xy[3], extent_xy[3], nstep)
        elif side == 'right':
            box_x = np.linspace(extent_xy[x1_index], extent_xy[x1_index], nstep)
            box_y = np.linspace(extent_xy[2], extent_xy[3], nstep)
        else:
            box_x = np.linspace(extent_xy[x1_index], extent_xy[x0_index], nstep)
            box_y = np.linspace(extent_xy[2], extent_xy[2], nstep)

        box_lon, box_lat = proj_inverse(box_x, box_y)
        box_x2, box_y2 = proj(box_lon, box_lat)
        use_box, = np.where(np.isclose(box_x2, box_x)
                            & np.isclose(box_y2, box_y))

        x = np.concatenate((bounds_x, box_x[use_box]))
        y = np.concatenate((bounds_y, box_y[use_box]))

        if side == 'left':
            st = np.argsort(y)
        elif side == 'top':
            st = np.argsort(x)
        elif side == 'right':
            st = np.argsort(y)[::-1]
        else:
            st = np.argsort(x)[::-1]

        return np.column_stack((x[st], y[st]))

    bounds_xy_clipped_sides.append(_generate_side('left'))
    bounds_xy_clipped_sides.append(_generate_side('top'))
    bounds_xy_clipped_sides.append(_generate_side('right'))
    bounds_xy_clipped_sides.append(_generate_side('bottom'))

    return np.concatenate(bounds_xy_clipped_sides)


def get_autoscale_vmin_vmax(values, vmin_in, vmax_in):
    """Get the autoscale vmin/vmax.

    Parameters
    ----------
    values : `np.ndarray`
        Array of values.
    vmin_in : `float` or `None`
        Input vmin (will not be overridden if set.)
    vmax_in : `float` or `None`
        Output vmin (will not be overridden if set.)

    Returns
    -------
    vmin, vmax : `float`
        Output vmin/vmax.
    """
    if values.dtype == bool or len(values) == 0:
        _vmin, _vmax = 0, 1
    else:
        _vmin, _vmax = np.nanpercentile(values, (AUTOSCALE_PERCENTILE_LOW, AUTOSCALE_PERCENTILE_HIGH))

    if _vmin == _vmax:
        # This will make the color scaling work decently well when we
        # have a flat integer type map.
        _vmin -= 0.1
        _vmax += 0.1

    vmin = vmin_in if vmin_in is not None else _vmin
    vmax = vmax_in if vmax_in is not None else _vmax

    return vmin, vmax



import math
import numpy as np

from pyproj import CRS
from pyproj import Transformer
from pyproj.exceptions import ProjError

from .utils import wrap_values

__all__ = ["SkyCRS", "PlateCarreeCRS", "McBrydeThomasFlatPolarQuarticCRS", "MollweideCRS",
           "HammerCRS", "EqualEarthCRS", "LambertAzimuthalEqualAreaCRS", "GnomonicCRS",
           "ObliqueMollweideCRS", "AlbersEqualAreaCRS", "get_crs", "get_available_crs"]


RADIUS = 1.0


class SkyCRS(CRS):
    """Coordinate Reference System (CRS) class describing sky projections.

    This is a specialized subclass of a `pyproj.CRS` Coordinate Reference
    System object.  Unlike a general earth-based CRS it uses a fixed
    reference radius instead of a generalized ellipsoidal model.  And it
    has additional routines used by SkyProj.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS type.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name=None, radius=RADIUS, **kwargs):
        self._name = name
        self.proj4_params = {'ellps': 'sphere',
                             'R': radius}
        self.proj4_params.update(**kwargs)
        super().__init__(self.proj4_params)

    def with_new_center(self, lon_0, lat_0=None):
        """Create a new SkyCRS with a new lon_0/lat_0.

        Parameters
        ----------
        lon_0 : `float`
            New longitude center.
        lat_0 : `float`, optional
            New latitude center (for projections that support it.)

        Returns
        -------
        crs : `skyproj.SkyCRS`
            New projection CRS.
        """
        proj4_params = self.proj4_params.copy()
        proj4_params['lon_0'] = lon_0
        if lat_0 is not None:
            proj4_params['lat_0'] = lat_0

        return self.__class__(**proj4_params)

    def transform_points(self, src_crs, x, y):
        """Transform points from a source coordinate reference system (CRS)
        to this CRS.

        Parameters
        ----------
        src_crs : `skyproj.SkyCRS`
            Source coordinate reference system describing x/y points.
        x : `np.ndarray`
            Array of x values, may be any number of dimensions.
        y : `np.ndarray`
            Array of y values, may be any number of dimensions.

        Returns
        -------
        result : `np.ndarray`
            Array of transformed points. Dimensions are [dim_x, 2]
            such that the final index gives the transformed x_prime, y_prime.
        """
        result_shape = tuple(x.shape[i] for i in range(x.ndim)) + (2, )

        x = x.ravel()
        y = y.ravel()

        npts = x.shape[0]

        result = np.zeros([npts, 2], dtype=np.float64)
        if npts:
            if isinstance(src_crs, PlateCarreeCRS):
                # We need to wrap to [-180, 180)
                x = wrap_values(x)
            try:
                transformer = Transformer.from_crs(src_crs, self, always_xy=True)
                result[:, 0], result[:, 1] = transformer.transform(x, y, None, errcheck=False)
            except ProjError as err:
                msg = str(err).lower()
                if (
                    "latitude" in msg
                    or "longitude" in msg
                    or "outside of projection domain" in msg
                    or "tolerance condition error" in msg
                ):
                    result[:] = np.nan
                else:
                    raise

            # and set to nans
            result[~np.isfinite(result)] = np.nan

        if len(result_shape) > 2:
            return result.reshape(result_shape)

        return result

    @property
    def lon_0(self):
        return self.proj4_params['lon_0']

    @property
    def lat_0(self):
        if 'lat_0' in self.proj4_params:
            return self.proj4_params['lat_0']
        else:
            return None

    @property
    def lat_1(self):
        if 'lat_1' in self.proj4_params:
            return self.proj4_params['lat_1']
        else:
            return None

    @property
    def lat_2(self):
        if 'lat_2' in self.proj4_params:
            return self.proj4_params['lat_2']
        else:
            return None

    @property
    def name(self):
        return self._name

    @property
    def radius(self):
        return self.proj4_params['R']

    def _as_mpl_transform(self, axes=None):
        from .transforms import SkyTransform

        if axes is None:
            raise ValueError("help need axes")

        return (SkyTransform(self) + axes.transData)

    def _as_mpl_axes(self):
        import matplotlib
        from .skyaxes import SkyAxes

        axes = SkyAxes

        # Support for old matplotlib.
        version_parts = matplotlib.__version__.split('.')
        if int(version_parts[0]) <= 3 and int(version_parts[1]) < 6:
            # Monkey-patch in the old cla.
            axes.cla = axes.cla_mplpre36

        return axes, {'sky_crs': self}


class PlateCarreeCRS(SkyCRS):
    """Equirectangular (Plate carree) sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``cyl``.
    lon_0 : `float`, optional
        Central longitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='cyl', lon_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'eqc',
                        'lon_0': lon_0,
                        'to_meter': math.radians(1)*radius,
                        'vto_meter': 1}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class McBrydeThomasFlatPolarQuarticCRS(SkyCRS):
    """McBryde Thomas Flat Polar Quartic sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``mbtfpq``.
    lon_0 : `float`, optional
        Central longitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='mbtfpq', lon_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'mbtfpq',
                        'lon_0': lon_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class MollweideCRS(SkyCRS):
    """Mollweide sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``moll``.
    lon_0 : `float`, optional
        Central longitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='moll', lon_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'moll',
                        'lon_0': lon_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class ObliqueMollweideCRS(SkyCRS):
    """Oblique Mollweide sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``moll``.
    lon_0 : `float`, optional
        Central longitude of projection.
    lat_p : `float`, optional
        Latitude of the North Pole of the unrotated coordinate system.
    lon_p : `float`, optional
        Longitude of the North Pole of the unrotated coordinate system.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='obmoll', lon_0=0.0, lat_p=90.0, lon_p=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'ob_tran',
                        'o_proj': 'moll',
                        'o_lat_p': lat_p,
                        'o_lon_p': lon_p,
                        'lon_0': lon_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)

    @property
    def lon_0(self):
        return self.proj4_params['lon_0'] + self.proj4_params['o_lon_p']


class HammerCRS(SkyCRS):
    """Hammer-Aitoff sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``hammer``.
    lon_0 : `float`, optional
        Central longitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='hammer', lon_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'hammer',
                        'lon_0': lon_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class EqualEarthCRS(SkyCRS):
    """Equal Earth sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``eqearth``.
    lon_0 : `float`, optional
        Central longitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='eqearth', lon_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'eqearth',
                        'lon_0': lon_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class LambertAzimuthalEqualAreaCRS(SkyCRS):
    """Lambert Azimuthal Equal Area sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``laea``.
    lon_0 : `float`, optional
        Central longitude of projection.
    lat_0 : `float`, optional
        Central latitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='laea', lon_0=0.0, lat_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'laea',
                        'lon_0': lon_0,
                        'lat_0': lat_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class GnomonicCRS(SkyCRS):
    """Gnomonic sky CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``gnom``.
    lon_0 : `float`, optional
        Central longitude of projection.
    lat_0 : `float`, optional
        Central latitude of projection.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='gnom', lon_0=0.0, lat_0=0.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'gnom',
                        'lon_0': lon_0,
                        'lat_0': lat_0}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


class AlbersEqualAreaCRS(SkyCRS):
    """Albers Equal Area CRS.

    Parameters
    ----------
    name : `str`, optional
        Name of projection CRS. Must be ``aea``.
    lon_0 : `float`, optional
        Central longitude of projection.
    lat_1 : `float`, optional
        First standard parallel.
    lat_2 : `float`, optional
        Second standard parallel.
    radius : `float`, optional
        Radius of projected sphere.
    **kwargs : `dict`, optional
        Additional kwargs for PROJ4 parameters.
    """
    def __init__(self, name='aea', lon_0=0.0, lat_1=15.0, lat_2=45.0, radius=RADIUS, **kwargs):
        proj4_params = {'proj': 'aea',
                        'lon_0': lon_0,
                        'lat_1': lat_1,
                        'lat_2': lat_2}
        proj4_params = {**proj4_params, **kwargs}

        super().__init__(name=name, radius=radius, **proj4_params)


_crss = {
    'hammer': ('Hammer', HammerCRS),
    'mbtfpq': ('McBryde-Thomas Flat Polar Quartic', McBrydeThomasFlatPolarQuarticCRS),
    'cyl': ('Plate Carree', PlateCarreeCRS),
    'eqearth': ('Equal Earth', EqualEarthCRS),
    'laea': ('Lambert Azimuthal Equal Area', LambertAzimuthalEqualAreaCRS),
    'moll': ('Mollweide', MollweideCRS),
    'obmoll': ('Oblique Mollweide', ObliqueMollweideCRS),
    'gnom': ('Gnomonic', GnomonicCRS),
    'aea': ('Albers Equal Area', AlbersEqualAreaCRS),
}


def get_crs(name, **kwargs):
    """Return a skyproj CRS.

    For list of projections available, use skyproj.get_available_crs().

    Parameters
    ----------
    name : `str`
        Skyproj name of projection CRS.
    **kwargs :
        Additional kwargs appropriate for given projection CRS.

    Returns
    -------
    crs : `skyproj.SkyCRS`
    """
    # Is this a listed projection CRS?
    if name not in _crss:
        raise ValueError(f'{name} CRS name is not recognized.  See get_available_crs()')

    descr, crsclass = _crss[name]

    return crsclass(name=name, **kwargs)


def get_available_crs():
    """Return dict of available projection CRSs.

    Returns
    -------
    available_crs: `dict`
        Available CRSs.  Key is skyproj name, value is brief description.
    """
    available_crs = {}
    for name, (descr, crsclass) in _crss.items():
        available_crs[name] = descr

    return available_crs



import os
import pytest

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize("lon_0", [0.0, 180.0])
def test_lines_polygons_mcbryde(tmp_path, lon_0):
    """Test drawing lines and polygons."""
    plt.rcParams.update(plt.rcParamsDefault)

    # This code draws a bunch of geodesics and polygons that are
    # both empty and filled, and cross over the boundary, to ensure
    # that features are working as intended.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax, lon_0=lon_0)

    # Draw two geodesics, one of which will wrap around.
    sp.plot([-10., 45.], [-10., 45.], 'r-', label='One')
    sp.plot([170., 210.], [-10., 45.], 'b--', label='Two')

    # Draw two unfilled polygons, one of which will wrap around.
    sp.draw_polygon([-20, 20, 20, -20], [20, 20, 40, 40],
                    edgecolor='magenta', label='Three')
    sp.draw_polygon([160, 200, 200, 160], [20, 20, 40, 40],
                    edgecolor='black', label='Four')

    # Draw two filled polygons, one of which will wrap around.
    sp.draw_polygon([-20, 20, 20, -20], [-20, -20, -40, -40],
                    edgecolor='black', facecolor='red', linestyle='--', label='Five')
    sp.draw_polygon([160, 200, 200, 160], [-20, -20, -40, -40],
                    edgecolor='red', facecolor='black', linestyle='-', label='Six')

    # Draw two circles, one empty, one filled.
    sp.circle(40.0, -40.0, 5.0, color='blue', label='Seven')
    sp.circle(-40.0, -40.0, 5.0, color='orange', label='Eight', fill=True)

    # Test ``ellipse``.  We can only plot one point per call
    sp.ellipse(60, 15, 10, 4, 0, color='green', label='Nine')
    sp.ellipse(300, 15, 15, 2, 45, fill=True, color='red', label='Ten')

    sp.legend()
    fname = f'lines_and_polygons_{lon_0}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("lonlatplonp", [(0.0, 45.0, -90.0),
                                         (100.0, 80.0, -90.0)])
def test_lines_polygons_obmoll(tmp_path, lonlatplonp):
    """Test drawing lines and polygons on an Oblique Mollweide map."""
    plt.rcParams.update(plt.rcParamsDefault)

    lon_0, lat_p, lon_p = lonlatplonp

    # This code draws a bunch of geodesics and polygons that are
    # both empty and filled, and cross over the boundary, to ensure
    # that features are working as intended.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.ObliqueMollweideSkyproj(ax=ax, lon_0=lon_0, lat_p=lat_p, lon_p=lon_p)

    # Draw two geodesics, one of which will wrap around.
    sp.plot([-10., 45.], [-10., 45.], 'r-', label='One')
    sp.plot([170., 210.], [-10., 45.], 'b--', label='Two')

    # Draw two unfilled polygons, one of which will wrap around.
    sp.draw_polygon([-20, 20, 20, -20], [20, 20, 40, 40],
                    edgecolor='magenta', label='Three')
    sp.draw_polygon([160, 200, 200, 160], [20, 20, 40, 40],
                    edgecolor='black', label='Four')

    # Draw two filled polygons, one of which will wrap around.
    # Note that wrapped filled polygons do not currently work with oblique transforms.
    sp.draw_polygon([-20, 20, 20, -20], [-20, -20, -40, -40],
                    edgecolor='black', facecolor='red', linestyle='--', label='Five')
    sp.draw_polygon([160, 200, 200, 160], [-20, -20, -40, -40],
                    edgecolor='red', facecolor='black', linestyle='-', label='Six')

    sp.legend()
    fname = f'lines_and_polygons_obmoll_{lon_0}_{lat_p}_{lon_p}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)



import numpy as np
import hpgeom as hpg

from .utils import wrap_values

__all__ = ['healpix_pixels_range', 'hspmap_to_xy', 'hpxmap_to_xy', 'healpix_to_xy',
           'healpix_bin']


def healpix_pixels_range(nside, pixels, wrap, nest=False):
    """Find lon/lat range of healpix pixels, using wrap angle.

    Parameters
    ----------
    nside : `int`
        Healpix nside
    pixels : `np.ndarray`
        Array of pixel numbers
    wrap : `float`
        Wrap angle.
    nest : `bool`, optional
        Nest ordering?

    Returns
    -------
    lon_range : `tuple` [`float`, `float`]
        Longitude range of pixels (min, max)
    lat_range : `tuple` [`float`, `float`]
        Latitude range of pixels (min, max)
    """
    lon, lat = hpg.pixel_to_angle(nside, pixels, nest=nest)

    eps = hpg.max_pixel_radius(nside)

    lat_range = (np.clip(np.min(lat) - eps, -90.0 + 1e-5, None),
                 np.clip(np.max(lat) + eps, None, 90.0 - 1e-5))

    # FIXME: the wrap logic here is wrong.

    lon_wrap = (lon + wrap) % 360. - wrap
    lon_range = (np.min(lon_wrap) - eps, np.max(lon_wrap) + eps)

    # Check if we have overrun and need to do the full range
    full_range = False
    if (lon_range[0] < (wrap - 360.0)) and (lon_range[1] > (wrap - 360.0)):
        full_range = True
    elif (lon_range[0] < (wrap + 360.0)) and (lon_range[1] > (wrap + 360.0)):
        full_range = True
    elif (lon_range[1] - lon_range[0]) >= 359.0:
        full_range = True

    if full_range:
        lon_0 = wrap_values((wrap + 180.0) % 360.0)
        lon_range = (lon_0 - 180., lon_0 + 180.0 - 1e-5)

    return lon_range, lat_range


def hspmap_to_xy(hspmap, lon_range, lat_range, xsize=1000, aspect=1.0, valid_mask=False):
    """Convert healsparse map to rasterized x/y positions and values.

    Parameters
    ----------
    hspmap : `healsparse.HealSparseMap`
        Healsparse map
    lon_range : `tuple` [`float`, `float`]
        Longitude range for rasterization, (min, max).
    lat_range : `tuple` [`float`, `float`]
        Latitude range for rasterization, (min, max).
    xsize : `int`, optional
        Number of rasterized pixels in the x direction.
    aspect : `float`, optional
        Aspect ratio for ysize.
    valid_mask : `bool`, optional
        Plot the valid pixels of the map.

    Returns
    -------
    lon_raster : `np.ndarray`
        Rasterized longitude values (length xsize).
    lat_raster : `np.ndarray`
        Rasterized latitude values (length xsize*aspect).
    values_raster : `np.ma.maskedarray`
        Rasterized values (2-d).  Invalid values are masked.
    """
    lon_raster, lat_raster = np.meshgrid(np.linspace(lon_range[0], lon_range[1], xsize),
                                         np.linspace(lat_range[0], lat_range[1], int(aspect*xsize)))

    # For pcolormesh we need the central locations
    clon = (lon_raster[1:, 1:] + lon_raster[:-1, :-1])/2.
    clat = (lat_raster[1:, 1:] + lat_raster[:-1, :-1])/2.

    values = hspmap.get_values_pos(clon, clat, valid_mask=valid_mask)
    if hspmap.is_wide_mask_map:
        # Special case wide masks.  We just display 1 where any bit
        # is defined, and 0 otherwise.
        values = np.any(values, axis=2).astype(np.int32)
        mask = (values == 0)
    elif values.dtype == bool:
        mask = values == hspmap._sentinel
    else:
        mask = (np.isclose(values, hspmap._sentinel) | np.isnan(values))

    return lon_raster, lat_raster, np.ma.array(values, mask=mask)


def hpxmap_to_xy(hpxmap, lon_range, lat_range, nest=False, xsize=1000, aspect=1.0):
    """Convert healpix map to rasterized x/y positions and values.

    Parameters
    ----------
    hpxmap : `np.ndarray`
        Healpix map
        lon_range : `tuple` [`float`, `float`]
        Longitude range for rasterization, (min, max).
    lat_range : `tuple` [`float`, `float`]
        Latitude range for rasterization, (min, max).
    xsize : `int`, optional
        Number of rasterized pixels in the x direction.
    aspect : `float`, optional
        Aspect ratio for ysize.

    Returns
    -------
    lon_raster : `np.ndarray`
        Rasterized longitude values (length xsize).
    lat_raster : `np.ndarray`
        Rasterized latitude values (length xsize*aspect).
    values_raster : `np.ma.maskedarray`
        Rasterized values (2-d).  Invalid values are masked.
    """
    lon_raster, lat_raster = np.meshgrid(np.linspace(lon_range[0], lon_range[1], xsize),
                                         np.linspace(lat_range[0], lat_range[1], int(aspect*xsize)))

    # For pcolormesh we need the central locations
    clon = (lon_raster[1:, 1:] + lon_raster[:-1, :-1])/2.
    clat = (lat_raster[1:, 1:] + lat_raster[:-1, :-1])/2.

    pix_raster = hpg.angle_to_pixel(hpg.npixel_to_nside(hpxmap.size), clon, clat, nest=nest)
    values = hpxmap[pix_raster]

    mask = (np.isclose(values, hpg.UNSEEN) | np.isnan(values))

    return lon_raster, lat_raster, np.ma.array(values, mask=mask)


def healpix_to_xy(nside, pixels, values, lon_range, lat_range,
                  nest=False, xsize=1000, aspect=1.0):
    """Convert healpix pixels to rasterized x/y positions and values.

    Parameters
    ----------
    nside : `int`
        Healpix nside.
    pixels : `np.ndarray`
        Array of pixel numbers
    values : `np.ndarray`
        Array of pixel values
    lon_range : `tuple`, optional
        Longitude range to do rasterization (min, max).
    lat_range : `tuple`, optional
        Latitude range to do rasterization (min, max).
    nest : `bool`, optional
        Nest ordering?
    xsize : `int`, optional
        Number of rasterized pixels in the x direction.
    aspect : `float`, optional
        Aspect ratio to compute ysize.

    Returns
    -------
    lon_raster : `np.ndarray`
        Rasterized longitude values (length xsize).
    lat_raster : `np.ndarray`
        Rasterized latitude values (length xsize*aspect).
    values_raster : `np.ndarray`
        Rasterized values (2-d).
    """
    test = np.unique(pixels)
    if test.size != pixels.size:
        raise ValueError("The pixels array must be unique.")

    lon_raster, lat_raster = np.meshgrid(np.linspace(lon_range[0], lon_range[1], xsize),
                                         np.linspace(lat_range[0], lat_range[1], int(aspect*xsize)))

    # For pcolormesh we need the central locations
    clon = (lon_raster[1:, 1:] + lon_raster[:-1, :-1])/2.
    clat = (lat_raster[1:, 1:] + lat_raster[:-1, :-1])/2.

    pix_raster = hpg.angle_to_pixel(nside, clon, clat, nest=nest)

    st = np.argsort(pixels)
    sub1 = np.searchsorted(pixels, pix_raster, sorter=st)

    if pix_raster.max() > pixels.max():
        bad = np.where(sub1 == pixels.size)
        sub1[bad] = pixels.size - 1

    sub2 = np.where(pixels[st[sub1]] == pix_raster)
    sub1 = st[sub1[sub2]]

    mask = np.ones(pix_raster.shape, dtype=bool)
    mask[sub2] = False
    values_raster = np.zeros(pix_raster.shape, dtype=values.dtype)
    values_raster[sub2] = values[sub1]

    return lon_raster, lat_raster, np.ma.array(values_raster, mask=mask)


def healpix_bin(lon, lat, C=None, nside=256, nest=False):
    """Create a healpix histogram of counts in lon/lat space.

    Parameters
    ----------
    lon : `np.ndarray`
        Longitude array (degrees).
    lat : `np.ndarray`
        Latitude array (degrees).
    C : `np.ndarray`, optional
        Array of values to take average, paired with lon/lat.
    nside : `int`, optional
        Healpix nside resolution.
    nest : `bool`, optional
        Map in nest format?

    Returns
    -------
    hpxmap : `np.ndarray`
        Healsparse map of counts.
    """
    pix = hpg.angle_to_pixel(nside, lon, lat, nest=nest)

    count = np.zeros(hpg.nside_to_npixel(nside), dtype=np.int32)
    np.add.at(count, pix, 1)
    good = count > 0

    if C is not None:
        hpxmap = np.zeros(hpg.nside_to_npixel(nside))
        np.add.at(hpxmap, pix, C)
        hpxmap[good] /= count[good]
    else:
        hpxmap = count.astype(np.float64)

    hpxmap[~good] = hpg.UNSEEN

    return hpxmap



import os
import pytest

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize("survey_tuple", [(skyproj.DESSkyproj, "DES"),
                                          (skyproj.BlissSkyproj, "BLISS"),
                                          (skyproj.MaglitesSkyproj, "MagLiTeS"),
                                          (skyproj.DecalsSkyproj, "DECaLS"),
                                          (skyproj.DESAlbersSkyproj, "DESAlbers")])
def test_survey_outlines(tmp_path, survey_tuple):
    """Test drawing survey outlines."""
    plt.rcParams.update(plt.rcParamsDefault)

    survey = survey_tuple[0]
    name = survey_tuple[1]

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = survey(ax=ax)
    if name == 'DES' or name == 'DESAlbers':
        sp.draw_des(label=name)
    elif name == 'BLISS':
        sp.draw_bliss(label=name)
    elif name == 'MagLiTeS':
        sp.draw_maglites(label=name)
    elif name == 'DECaLS':
        sp.draw_decals(label=name)
    sp.legend()
    fname = f'{name}_survey.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)



import os
import pytest
import numpy as np
import healsparse as hsp

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize("skyproj", [skyproj.Skyproj,
                                     skyproj.McBrydeSkyproj,
                                     skyproj.MollweideSkyproj,
                                     skyproj.HammerSkyproj,
                                     skyproj.EqualEarthSkyproj,
                                     skyproj.AlbersSkyproj])
@pytest.mark.parametrize("lon_0", [0.0, -100.0, 100.0, 180.0])
def test_skyproj_basic(tmp_path, skyproj, lon_0):
    """Test full sky maps."""
    plt.rcParams.update(plt.rcParamsDefault)

    # Full image
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax, lon_0=lon_0)
    fname = f'{sp.projection_name}_full_{lon_0}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("skyproj", [skyproj.Skyproj,
                                     skyproj.McBrydeSkyproj,
                                     skyproj.MollweideSkyproj,
                                     skyproj.HammerSkyproj,
                                     skyproj.EqualEarthSkyproj,
                                     skyproj.LaeaSkyproj,
                                     skyproj.AlbersSkyproj])
def test_skyproj_zoom(tmp_path, skyproj):
    plt.rcParams.update(plt.rcParamsDefault)

    # Simple zoom
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax, extent=[0, 50, 0, 50])
    fname = f'{sp.projection_name}_zoom.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("lonlat", [(0.0, 0.0),
                                    (120.0, -75.0),
                                    (-120.0, 75.0)])
def test_skyproj_gnom(tmp_path, lonlat):
    """Test gnomonic zooms."""
    plt.rcParams.update(plt.rcParamsDefault)

    lon_0, lat_0 = lonlat

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.GnomonicSkyproj(ax=ax, lon_0=lon_0, lat_0=lat_0)
    # draw a square square, make sure it looks square
    delta_lat = 0.1
    delta_lon = delta_lat/np.cos(np.deg2rad(lat_0))
    sp.draw_polygon(
        [lon_0 - delta_lon, lon_0 + delta_lon, lon_0 + delta_lon, lon_0 - delta_lon],
        [lat_0 - delta_lat, lat_0 - delta_lat, lat_0 + delta_lat, lat_0 + delta_lat]
    )
    fname = f'gnom_{lon_0}_{lat_0}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("lonlatplonp", [(0.0, 45.0, -90.0),
                                         (100.0, 80.0, 0.0)])
def test_skyproj_obmoll(tmp_path, lonlatplonp):
    """Test Oblique Mollweide."""
    plt.rcParams.update(plt.rcParamsDefault)

    lon_0, lat_p, lon_p = lonlatplonp

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.ObliqueMollweideSkyproj(ax=ax, lon_0=lon_0, lat_p=lat_p, lon_p=lon_p)
    fname = f'{sp.projection_name}_{lon_0}_{lat_p}_{lon_p}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("lat1lat2", [(15.0, 45.0),
                                      (-20.0, 15.0),
                                      (-15.0, -45.0)])
def test_skyproj_albers(tmp_path, lat1lat2):
    """Test Albers Equal Area."""
    plt.rcParams.update(plt.rcParamsDefault)

    lat_1, lat_2 = lat1lat2

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.AlbersSkyproj(ax=ax, lat_1=lat_1, lat_2=lat_2)
    fname = f'{sp.projection_name}_{lat_1}_{lat_2}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("skyproj", [skyproj.Skyproj,
                                     skyproj.McBrydeSkyproj,
                                     skyproj.MollweideSkyproj,
                                     skyproj.HammerSkyproj,
                                     skyproj.EqualEarthSkyproj])
@pytest.mark.parametrize("lon_0", [-180.0, -140.0, -100.0, -60.0, -20.0, 0.0,
                                   20.0, 60.0, 100.0, 140.0, 180.0])
def test_skyproj_fullsky_extent(skyproj, lon_0):
    """Test getting the full sky extent."""

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax, lon_0=lon_0)

    lon1, lon0, lat0, lat1 = sp.get_extent()

    # We allow some slop because of the way that clipping works
    assert lat0 < (-90.0 + sp._pole_clip + 0.3)
    assert lat0 >= -90.0
    assert lat1 > (90.0 - sp._pole_clip - 0.3)
    assert lat1 <= 90.0

    # Cover (almost) full 360 deg?
    assert (lon1 - lon0) > (360.0 - 1.5)
    # And the rotated start is close to -180?
    rot = (lon0 - lon_0 + 180.0) % 360.0 - 180.0
    assert rot < (-180.0 + 1.0)


@pytest.mark.parametrize("skyproj", [skyproj.McBrydeSkyproj,
                                     skyproj.MollweideSkyproj])
def test_skyproj_nogap_180(tmp_path, skyproj):
    """Test that there is no mysterious gap when lon_0=180.0"""
    plt.rcParams.update(plt.rcParamsDefault)

    testmap = hsp.HealSparseMap.make_empty(32, 256, dtype=np.int32)
    poly = hsp.geom.Polygon(ra=[-20, 20, 20, -20], dec=[-20, -20, 20, 20], value=1)
    pixels = poly.get_pixels(nside=testmap.nside_sparse)
    testmap[pixels] = 1
    poly = hsp.geom.Polygon(ra=[160, 200, 200, 160], dec=[-20, -20, 20, 20], value=1)
    pixels = poly.get_pixels(nside=testmap.nside_sparse)
    testmap[pixels] = 1

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax, lon_0=180.0)
    sp.draw_hspmap(testmap, zoom=False)
    fname = f'{sp.projection_name}_gaptest.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)


def test_skyproj_override_sizes(tmp_path):
    """Test overriding the label/width sizes."""
    plt.rcParams.update(plt.rcParamsDefault)

    rcparams = {'xtick.labelsize': 20,
                'ytick.labelsize': 4,
                'axes.linewidth': 5}

    # Full image
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    _ = skyproj.McBrydeSkyproj(ax=ax, rcparams=rcparams)
    fname = 'skyproj_full_override_sizes.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)

    # And confirm that the changes do not carry over to another plot.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    _ = skyproj.McBrydeSkyproj(ax=ax)
    fname = 'mbtfpq_full_0.0.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 15.0)
    if err:
        raise ImageComparisonFailure(err)



import pytest

import skyproj


@pytest.mark.parametrize("crsname", ['hammer',
                                     'mbtfpq',
                                     'cyl',
                                     'eqearth',
                                     'laea',
                                     'moll',
                                     'gnom',
                                     'obmoll'])
def test_get_crs(crsname):
    """Test getting a projection CRS and making a new instance."""
    crs = skyproj.get_crs(crsname)

    crs2 = crs.__class__(**crs.proj4_params)

    for key in crs.proj4_params.keys():
        assert crs2.proj4_params[key] == crs.proj4_params[key]


@pytest.mark.parametrize("crsname", ['hammer',
                                     'mbtfpq',
                                     'cyl',
                                     'eqearth',
                                     'laea',
                                     'moll',
                                     'gnom',
                                     'obmoll'])
def test_update_crs_lon_0(crsname):
    """Test updating a projection CRS lon_0."""
    crs = skyproj.get_crs(crsname)

    crs2 = crs.with_new_center(100.0)

    assert crs.__class__ == crs2.__class__

    for key in crs.proj4_params.keys():
        if key == 'lon_0':
            assert crs2.lon_0 == 100.0
        else:
            assert crs2.proj4_params[key] == crs.proj4_params[key]


@pytest.mark.parametrize("crsname", ['laea',
                                     'gnom'])
def tets_update_crs_lon_0_lat_0(crsname):
    """Test updating a projection CRS lon_0/lat_0."""
    crs = skyproj.get_crs(crsname)

    crs2 = crs.with_new_center(100.0, lat_0=-45.0)

    assert crs.__class__ == crs2.__class__

    for key in crs.proj4_params.keys():
        if key == 'lon_0':
            assert crs2.lon_0 == 100.0
        elif key == 'lat_0':
            assert crs2.lat_0 == -45.0
        else:
            assert crs2.proj4_params[key] == crs.proj4_params[key]



def add_func_docstr(docstring):
    def _doc(func):
        func.__doc__ = docstring
        return func
    return _doc


def skyproj_init_parameters(intro_string, include_projection_name=False, include_lon_0=True):
    par_str = intro_string

    par_str += """

Parameters
----------
ax : `matplotlib.axes.Axes`, optional
    Axis object to replace with a skyproj axis."""

    if include_projection_name:
        par_str += """
projection_name : `str`, optional
    Valid proj4/cartosky projection name."""

    if include_lon_0:
        par_str += """
lon_0 : `float`, optional
    Central longitude of projection."""

    par_str += """
gridlines : `bool`, optional
    Draw gridlines?
celestial : `bool`, optional
    Do celestial plotting (e.g. invert longitude axis).
extent : iterable, optional
    Default exent of the map, [lon_min, lon_max, lat_min, lat_max].
    Note that lon_min, lon_max can be specified in any order, the
    orientation of the map is set by ``celestial``.
longitude_ticks : `str`, optional
    Label longitude ticks from 0 to 360 degrees (``positive``) or
    from -180 to 180 degrees (``symmetric``).
autorescale : `bool`, optional
    Automatically rescale color bars when zoomed?
galactic : `bool`, optional
    Plotting in Galactic coordinates?  Recommendation for Galactic plots
    is to have longitude_ticks set to ``symmetric`` and celestial = True.
rcparams : `dict`, optional
    Dictionary of matplotlib rc parameters to override.  In particular, the code will
    use ``xtick.labelsize`` and ``ytick.labelsize`` for the x and y tick labels, and
    ``axes.linewidth`` for the map boundary.
n_grid_lon : `int`, optional
    Number of gridlines to use in the longitude direction
    (default is axis_ratio * n_grid_lat).
n_grid_lat : `int`, optional
    Number of gridlines to use in the latitude direction (default is 6).
min_lon_ticklabel_delta : `float`, optional
    Minimum relative spacing between longitude tick labels (relative to width
    of axis). Smaller values yield closer tick labels (and potential for clashes)
    and larger values yield further tick labels."""

    return par_str


skyproj_kwargs_par = """
**kwargs : `dict`, optional
    Additional arguments to send to cartosky/proj4 projection CRS initialization.
"""


draw_map_common_pars = """
zoom : `bool`, optional
    Optimally zoom in projection to computed map.
xsize : `int`, optional
    Number of rasterized pixels in the x direction.
vmin : `float`, optional
    Minimum value for color scale.  Defaults to 2.5th percentile.
vmax : `float`, optional
    Maximum value for color scale.  Defaults to 97.5th percentile.
rasterized : `bool`, optional
    Plot with rasterized graphics.
lon_range : `tuple` [`float`, `float`], optional
    Longitude range to plot [``lon_min``, ``lon_max``].
lat_range : `tuple` [`float`, `float`], optional
    Latitude range to plot [``lat_min``, ``lat_max``].
**kwargs : `dict`
    Additional args to pass to pcolormesh."""


draw_map_returns = """

Returns
-------
im : `matplotlib.collections.QuadMesh`
    Image that was displayed
lon_raster : `np.ndarray`
    2D array of rasterized longitude values.
lat_raster : `np.ndarray`
    2D array of rasterized latitude values.
values_raster : `np.ma.MaskedArray`
    Masked array of rasterized values.
"""

draw_hpxmap_docstr = """Use pcolormesh to draw a healpix map.

Parameters
----------
hpxmap : `np.ndarray`
    Healpix map to plot, with length 12*nside*nside and UNSEEN for
    illegal values.
nest : `bool`, optional
    Map in nest ordering?""" + draw_map_common_pars + draw_map_returns


draw_hpxpix_docstr = """Use pcolormesh to draw a healpix map made of pixels and values.

Parameters
----------
nside : `int`
    Healpix nside of pixels to plot.
pixels : `np.ndarray`
    Array of pixels to plot.
values : `np.ndarray`
    Array of values associated with pixels.""" + draw_map_common_pars + draw_map_returns


draw_hspmap_docstr = """Use pcolormesh to draw a healsparse map.

Parameters
----------
hspmap : `healsparse.HealSparseMap`
    Healsparse map to plot.""" + draw_map_common_pars + draw_map_returns


draw_hpxbin_docstr = """Create a healpix histogram of counts in lon, lat.

Related to ``hexbin`` from matplotlib.

If ``C`` array is specified then the mean is taken from the C values.

Parameters
----------
lon : `np.ndarray`
    Array of longitude values.
lat : `np.ndarray`
    Array of latitude values.
C : `np.ndarray`, optional
    Array of values to average in each pixel.""" + draw_map_common_pars + draw_map_returns
